import numpy as np
import pandas as pd

import matplotlib.pyplot as plt

from sklearn.model_selection import (
    KFold,
    train_test_split,
    GridSearchCV,
    StratifiedKFold,
    cross_val_score,
    RandomizedSearchCV,
)
from sklearn.metrics import (
    precision_score,
    recall_score,
    f1_score,
    explained_variance_score,
    mean_squared_log_error,
    mean_absolute_error,
    median_absolute_error,
    mean_squared_error,
    r2_score,
    confusion_matrix,
    roc_curve,
    accuracy_score,
    roc_auc_score,
    homogeneity_score,
    completeness_score,
    classification_report,
    silhouette_samples,
)

from prettierplot.plotter import PrettierPlot
from prettierplot import style

import shap


def single_shap_value_tree(self, obs_ix, model, data):
    """
    Documentation:

        ---
        Description:
            Generate elements necessary for creating a SHAP force plot for a single
            observation. Works with tree_based models, including:
                - RandomForestClassifier (package: sklearn)
                - GradientBoostingClassifier (package: sklearn)
                - ExtraTreesClassifier (package: sklearn)
                - LGBMClassifier (package: lightgbm)
                - XGBClassifier (package: xgboost)

        ---
        Parameters:
            obs_ix : int
                Index of observation to analyze.
            model : model object
                Instantiated model object.
            data : Pandas DataFrame
                Dataset from which to slice individual observation's features.

        ---
        Returns:
            obs_data : array
                Feature values for the specified observation.
            base_value : float
                Expected prediction value.
            obs_shap_values : array
                Data array containing the SHAP values for the specified
                observation.
    """
    # collect observation feature values, explainer object and observation SHAP values
    obs_data = data.loc[obs_ix].values.reshape(1, -1)
    explainer = SHAP.TreeExplainer(model.model)
    obs_shap_values = explainer.shap_values(obs_data)

    # accommodate the fact that different types of models generate differently
    # formatted SHAP values and expected values
    if isinstance(obs_shap_values, list):
        obs_shap_values = obs_shap_values[1]
    else:
        obs_shap_values = obs_shap_values

    if isinstance(explainer.expected_value, np.floating):
        base_value = explainer.expected_value
    else:
        base_value = explainer.expected_value[1]

    return obs_data, base_value, obs_shap_values


def single_shap_viz_tree(self, obs_ix, model, data, target=None, classification=True, cmap="viridis"):
    """
    Documentation:

        ---
        Description:
            Generate a SHAP force plot for a single observation. Works with tree-based models,
            including:
                - RandomForestClassifier (package: sklearn)
                - GradientBoostingClassifier (package: sklearn)
                - ExtraTreesClassifier (package: sklearn)
                - LGBMClassifier (package: lightgbm)
                - XGBClassifier (package: xgboost)

        ---
        Parameters:
            obs_ix : int
                Index of observation to visualize.
            model : model object
                Instantiated model object.
            data : Pandas DataFrame
                Dataset from which to slice indiviudal observation's features.
            target : Pandas Series, default=None
                True label for observation.
            classification : bool, default=True
                Bool argument indicating whether the supervised learning task is classification
                task or regression task.
            cmap : str, colormap, default=viridis
                Colormap applied to plot.
    """
    # return observation features values, expected value, and observation SHAP values
    obs_data, base_value, obs_shap_values = self.single_shap_value_tree(obs_ix=obs_ix, model=model, data=data)

    # display summary information about prediction
    if classification:
        probas = model.predict_proba(obs_data)
        print('negative class probability: {:.6f}'.format(probas[0][0]))
        print('positive class probability: {:.6f}'.format(probas[0][1]))

        if target is not None:
            print('True label: {}'.format(target.loc[obs_ix]))
    else:
        print('prediction: {:.6f}'.format(model.predict(obs_data)[0]))
        if target is not None:
            print('True label: {:.6f}'.format(target.loc[obs_ix]))

    # display force plot
    SHAP.force_plot(
        base_value=base_value,
        shap_values=np.around(obs_shap_values.astype(np.double),3),
        features=np.around(obs_data.astype(np.double),3),
        feature_names=data.columns.tolist(),
        matplotlib=True,
        show=False,
        plot_cmap=cmap
    )
    plt.rcParams['axes.facecolor']='white'
    plt.rcParams['figure.facecolor'] = 'white'
    plt.grid(b=False)
    plt.show()


def multi_shap_value_tree(self, obs_ixs, model, data):
    """
    Documentation:

        ---
        Description:
            Generate elements necessary for creating a SHAP force plot for multiple
            observations simultaneously. Works with tree-based models, including:
                - RandomForestClassifier (package: sklearn)
                - GradientBoostingClassifier (package: sklearn)
                - ExtraTreesClassifier (package: sklearn)
                - LGBMClassifier (package: lightgbm)
                - XGBClassifier (package: xgboost)

        ---
        Parameters:
            obs_ixs : array or list
                Index values of observation to be analyze.
            model : model object
                Instantiated model object.
            data : Pandas DataFrame
                Dataset from which to slice indiviudal observations' feature values.
        
        ---
        Returns:
            obs_data : array
                Feature values for the specified observations.
            base_value : float
                Expected predicted value.
            obs_shap_values : array
                data array containing the SHAP values for the specified
                observations.
    """

    # collect observation feature values, explainer object and observation SHAP values
    obs_data = data.loc[obs_ixs].values
    explainer = SHAP.TreeExplainer(model.model)
    obs_shap_values = explainer.shap_values(obs_data)

    # accommodate the fact that different types of models generate differently
    # formatted SHAP values and expected values
    if isinstance(obs_shap_values, list):
        obs_shap_values = obs_shap_values[1]
    else:
        obs_shap_values = obs_shap_values

    if isinstance(explainer.expected_value, np.floating):
        base_value = explainer.expected_value
    else:
        base_value = explainer.expected_value[1]

    return obs_data, base_value, obs_shap_values


def multi_shap_viz_tree(self, obs_ixs, model, data):
    """
    Documentation:
        
        ---
        Description:
            Generate a SHAP force plot for multiple observations simultaneously. Works with
            tree-based models, including:
                - RandomForestClassifier (package: sklearn)
                - GradientBoostingClassifier (package: sklearn)
                - ExtraTreesClassifier (package: sklearn)
                - LGBMClassifier (package: lightgbm)
                - XGBClassifier (package: xgboost)

        ---
        Parameters:
            obs_ixs : array or list
                Index values of observations to analyze.
            model : model object
                Instantiated model object.
            data : Pandas DataFrame
                Dataset from which to slice observations' feature values.
    """
    # return observation features values, expected value, and observation SHAP values
    obs_data, base_value, obs_shap_values = self.multi_shap_value_tree(obs_ixs=obs_ixs, model=model, data=data)

    # generate force plot
    visual = SHAP.force_plot(
        base_value = base_value,
        shap_values = obs_shap_values,
        features = obs_data,
        feature_names = data.columns.tolist(),
    )
    return visual


def shap_dependence_plot(self, obs_data, obs_shap_values, scatter_feature, color_feature, feature_names,
                        x_jitter=0.08, dot_size=25, alpha=0.7, show=True, ax=None):
    """
    Documentation:
        
        ---
        Description:
            Generate a SHAP dependence plot for a pair of features. One feature is
            represented in a scatter plot, with each observation's actual value on the
            x-axis and the corresponding SHAP value for the observation on the y-axis.
            the second feature applies a hue to the scattered feature based on the
            individual values of the second feature.
        
        ---
        Parameters:
            obs_data : array
                Feature values for the specified observations.
            obs_shap_values : array
                Data array containing the SHAP values for the specified observations.
            scatter_feature : str
                Name of feature to scatter on plot area.
            color_feature : str
                Name of feature to use when applying color to dots on scatter plot.
            feature_names : list
                List of all feature names in the dataset.
            x_jitter : float, default=0.08
                Controls displacement of dots along x-axis.
            dot_size : float, default=25
                Size of dots.
            alpha : float, default=0.7
                Transparency of dots.
            show : boolean, default=True
                Conditional controlling whether to display plot automatically when
                calling function.
            ax : axes object, default=None
                Axis on which to place visual.
    """
    # generate force plot
    SHAP.dependence_plot(
            ind = scatter_feature,
            shap_values = obs_shap_values,
            features = obs_data,
            feature_names = feature_names,
            interaction_index = color_feature,
            show=False,
            x_jitter=x_jitter,
            dot_size=dot_size,
            alpha=alpha,
            ax=ax,
        )
    plt.rcParams['axes.facecolor'] = 'white'
    plt.rcParams['figure.facecolor'] = 'white'
    plt.grid(b=False)

    if show:
        plt.show()


def shap_dependence_grid(self, obs_data, obs_shap_values, grid_features, all_features, dot_size=25, alpha=0.7):
    """
    Documentation:
        Description:
            Generate a SHAP dependence plot for a pair of features. One feature is
            represented in a scatter plot, with each observation's actual value on the
            x-axis and the corresponding SHAP value for the observation on the y-axis.
            The second feature applies a hue to the scattered feature based on the
            individual values of the second feature.
        Parameters:
            obs_data : array
                feature values for the specified observations.
            obs_shap_values : array
                Data array containing the SHAP values for the specified
                observations.
            grid_features : list
                Names of features to display on grid.
            all_features : list
                List containing names for all features for which SHAP values were
                calculated.
            dot_size : float, default=25
                Size of dots.
            alpha : float, default=0.7
                Transparency of dots.
    """
    # create subplot grid    
    fig, ax = plt.subplots(
        ncols=len(grid_features),
        nrows=len(grid_features),
        constrained_layout=True,
        figsize=(len(grid_features) * 3.5, len(grid_features) * 2.5),
    )

    # iterate through column feature and row feature pairs
    for col_ix, col_feature in enumerate(grid_features):
        for row_ix, row_feature in enumerate(grid_features):
            self.shap_dependence_plot(
                obs_data=obs_data,
                obs_shap_values=obs_shap_values,
                scatter_feature=col_feature,
                color_feature=row_feature,
                feature_names=all_features,
                dot_size=dot_size,
                alpha=alpha,
                show=False,
                ax=ax[row_ix, col_ix],
            )
            ax[row_ix, col_ix].yaxis.set_visible(False)
            if row_ix != col_ix:
                ax[row_ix, col_ix].set_xlabel("{},\nby {}".format(col_feature, row_feature))
    plt.show()


def shap_summary_plot(self, obs_data, obs_shap_values, feature_names, alpha=0.7):
    """
    Documentation:
        Description:
            Generate a SHAP summary plot for multiple observations.
        Parameters:
            obs_data : array
                Feature values for the specified observations.
            obs_shap_values : array
                Data array containing the SHAP values for the specified
                observations.
            feature_names : list
                List of all feature names in the dataset.
            alpha : float, default=0.7
                Controls transparency of dots.
    """
    # generate SHAP summary plot
    SHAP.summary_plot(
        shap_values = obs_shap_values,
        features = obs_data,
        feature_names = feature_names,
        alpha = 0.7,
        show=False,
    )
    plt.rcParams['axes.facecolor'] = 'white'
    plt.grid(b=False)
    plt.show()


import numpy as np
import pandas as pd

import matplotlib.pyplot as plt

from sklearn.model_selection import (
    KFold,
    train_test_split,
    GridSearchCV,
    StratifiedKFold,
    cross_val_score,
    RandomizedSearchCV,
)
from sklearn.metrics import (
    precision_score,
    recall_score,
    f1_score,
    explained_variance_score,
    mean_squared_log_error,
    mean_absolute_error,
    median_absolute_error,
    mean_squared_error,
    r2_score,
    confusion_matrix,
    roc_curve,
    accuracy_score,
    roc_auc_score,
    homogeneity_score,
    completeness_score,
    classification_report,
    silhouette_samples,
    plot_confusion_matrix,
)

from scipy import stats

from prettierplot.plotter import PrettierPlot
from prettierplot import style


def binary_classification_panel(self, model, X_train, y_train, X_valid=None, y_valid=None, labels=None,
                        n_folds=None, title_scale=1.0, color_map="viridis", random_state=1, chart_scale=15):
    """
    Documentation:

        ---
        Description:
            Generate a panel of reports and visualizations summarizing the
            performance of a classification model.

        ---
        Parameters:
            model : model object
                Instantiated model object.
            X_train : Pandas DataFrame
                Training data observations.
            y_train : Pandas Series
                Training target data.
            X_valid : Pandas DataFrame, default=None
                Validation data observations.
            y_valid : Pandas Series, default=None
                Validation target data.
            labels : list, default=None
                Custom labels for confusion matrix axes. If left as none,
                will default to 0, 1, 2...
            n_folds : int, default=None
                Number of cross-validation folds to use. If validation data is provided through
                X_valid/y_valid, n_folds is ignored.
            color_map : str specifying built-in matplotlib colormap, default="viridis"
                Color map applied to plots.
            title_scale : float, default=1.0
                Controls the scaling up (higher value) and scaling down (lower value) of the size
                of the main chart title, the x_axis title and the y_axis title.
            random_state : int, default=1
                Random number seed.
            chart_scale : int or float, default=15
                Controls size and proportions of chart and chart elements. Higher value creates
                larger plots and increases visual elements proportionally.
    """
    print("*" * 55)
    print("* Estimator: {}".format(model.estimator_name))
    print("* Parameter set: {}".format(model.model_iter))
    print("*" * 55)

    print("\n" + "*" * 55)
    print("Training data evaluation\n")

    ## training panel
    # fit model on training data and generate predictions using training data
    y_pred = model.fit(X_train, y_train).predict(X_train)

    # print and generate classification_report using training data
    print(
            classification_report(
                y_train,
                y_pred,
                target_names=labels if labels is not None else np.unique(y_train.values),
            )
        )

    # create prettierplot object
    p = PrettierPlot(chart_scale=chart_scale, plot_orientation="wide_narrow")

    # add canvas to prettierplot object
    ax = p.make_canvas(
        title="Confusion matrix - training data\nModel: {}\nParameter set: {}".format(
            model.estimator_name, model.model_iter
        ),
        y_shift=0.4,
        x_shift=0.25,
        position=121,
        title_scale=title_scale,
    )

    # add confusion plot to canvas
    plot_confusion_matrix(
        estimator=model,
        X=X_train,
        y_true=y_train,
        display_labels=labels if labels is not None else np.unique(y_train.values),
        cmap=color_map,
        values_format=".0f",
        ax=ax,
    )

    # add canvas to prettierplot object
    ax = p.make_canvas(
        title="ROC curve - training data\nModel: {}\nParameter set: {}".format(
            model.estimator_name,
            model.model_iter,
        ),
        x_label="False positive rate",
        y_label="True positive rate",
        y_shift=0.35,
        position=122,
        title_scale=title_scale,
    )
    # add ROC curve to canvas
    p.roc_curve_plot(
        model=model,
        X_train=X_train,
        y_train=y_train,
        linecolor=style.style_grey,
        ax=ax,
    )
    plt.subplots_adjust(wspace=0.3)
    plt.show()

    # if validation data is provided
    if X_valid is not None:
        print("\n" + "*" * 55)
        print("Validation data evaluation\n")

        # fit model on training data and generate predictions using validation data
        y_pred = model.fit(X_train, y_train).predict(X_valid)

        # print and generate classification_report using training data
        print(
            classification_report(
                y_valid,
                y_pred,
                target_names=labels if labels is not None else np.unique(y_train.values),
            )
        )

        # create prettierplot object
        p = PrettierPlot(chart_scale=chart_scale, plot_orientation="wide_narrow")

        # add canvas to prettierplot object
        ax = p.make_canvas(
            title="Confusion matrix - validation data\nModel: {}\nParameter set: {}".format(
                model.estimator_name, model.model_iter
            ),
            y_shift=0.4,
            x_shift=0.25,
            position=121,
            title_scale=title_scale,
        )

        # add confusion matrix to canvas
        plot_confusion_matrix(
            estimator=model,
            X=X_valid,
            y_true=y_valid,
            display_labels=labels if labels is not None else np.unique(y_train.values),
            cmap=color_map,
            values_format=".0f",
            ax=ax,
        )

        # add canvas to prettierplot object
        ax = p.make_canvas(
            title="ROC curve - validation data\nModel: {}\nParameter set: {}".format(
                model.estimator_name,
                model.model_iter,
            ),
            x_label="False positive rate",
            y_label="True positive rate",
            y_shift=0.35,
            position=122,
            # position=111 if X_valid is not None else 121,
            title_scale=title_scale,
        )
        # add ROC curve to canvas
        p.roc_curve_plot(
            model=model,
            X_train=X_train,
            y_train=y_train,
            X_valid=X_valid,
            y_valid=y_valid,
            linecolor=style.style_grey,
            ax=ax,
        )
        plt.subplots_adjust(wspace=0.3)
        plt.show()

    # if n_folds are provided, indicating cross-validation
    elif isinstance(n_folds, int):
        print("\n" + "*" * 55)
        print("Cross validation evaluation\n")

        # generate cross-validation indices
        cv = list(
            StratifiedKFold(
                n_splits=n_folds, shuffle=True, random_state=random_state
            ).split(X_train, y_train)
        )

        # generate colors
        color_list = style.color_gen(color_map, num=len(cv))

        # iterate through cross-validation indices
        for i, (train_ix, valid_ix) in enumerate(cv):
            print("\n" + "*" * 55)
            print("CV Fold {}\n".format(i + 1))

            X_train_cv = X_train.iloc[train_ix]
            y_train_cv = y_train.iloc[train_ix]
            X_valid_cv = X_train.iloc[valid_ix]
            y_valid_cv = y_train.iloc[valid_ix]

            # fit model on training data and generate predictions using holdout observations
            y_pred = model.fit(X_train_cv, y_train_cv).predict(X_valid_cv)

            # print and generate classification_report using holdout observations
            print(
            classification_report(
                    y_valid_cv,
                    y_pred,
                    target_names=labels if labels is not None else np.unique(y_train.values),
                )
            )

            # create prettierplot object
            p = PrettierPlot(chart_scale=chart_scale, plot_orientation="wide_narrow")

            # add canvas to prettierplot object
            ax = p.make_canvas(
                title="Confusion matrix - CV Fold {}\nModel: {}\nParameter set: {}".format(
                    i + 1, model.estimator_name, model.model_iter
                ),
                y_shift=0.4,
                x_shift=0.25,
                position=121,
                title_scale=title_scale,
            )

            # add confusion matrix to canvas
            plot_confusion_matrix(
                estimator=model,
                X=X_valid_cv,
                y_true=y_valid_cv,
                display_labels=labels if labels is not None else np.unique(y_train.values),
                cmap=color_map,
                values_format=".0f",
                ax=ax,
            )

            # add canvas to prettierplot object
            ax = p.make_canvas(
                title="ROC curve - CV Fold {}\nModel: {}\nParameter set: {}".format(
                    i + 1,
                    model.estimator_name,
                    model.model_iter,
                ),
                x_label="False positive rate",
                y_label="True positive rate",
                y_shift=0.35,
                position=122,
                title_scale=title_scale,
            )

            # add ROC curve to canvas
            p.roc_curve_plot(
                model=model,
                X_train=X_train_cv,
                y_train=y_train_cv,
                X_valid=X_valid_cv,
                y_valid=y_valid_cv,
                linecolor=style.style_grey,
                ax=ax,
            )
            plt.subplots_adjust(wspace=0.3)
            plt.show()

def regression_panel(self, model, X_train, y_train, X_valid=None, y_valid=None, n_folds=None, title_scale=1.0,
                    color_map="viridis", random_state=1, chart_scale=15):
    """
    Documentation:
        Description:
            creates a set of residual plots and pandas DataFrames, where each row captures various summary statistics
            pertaining to a model's performance. generates residual plots and captures performance data for training
            and validation datasets. If no validation set is provided, then cross_validation is performed on the
            training dataset.
        Parameters:
            model : model object
                Instantiated model object.
            X_train : Pandas DataFrame
                Training data observations.
            y_train : Pandas Series
                Training target data.
            X_valid : Pandas DataFrame, default=None
                Validation data observations.
            y_valid : Pandas Series, default=None
                Validation target data.
            n_folds : int, default=None
                Number of cross-validation folds to use. If validation data is provided through
                X_valid/y_valid, n_folds is ignored.
            title_scale : float, default=1.0
                Controls the scaling up (higher value) and scaling down (lower value) of the size of
                the main chart title, the x_axis title and the y_axis title.
            color_map : str specifying built-in matplotlib colormap, default="viridis"
                Color map applied to plots.
            random_state : int, default=1
                Random number seed.
            chart_scale : int or float, default=15
                Controls size and proportions of chart and chart elements. Higher value creates larger plots
                and increases visual elements proportionally.
    """

    print("*" * 55)
    print("* Estimator: {}".format(model.estimator_name))
    print("* Parameter set: {}".format(model.model_iter))
    print("*" * 55)

    print("\n" + "*" * 55)
    print("Training data evaluation")

    # fit model on training data
    model.fit(X_train.values, y_train.values)

    ## training dataset
    # generate predictions using training data and calculate residuals
    y_pred = model.predict(X_train.values)
    residuals = y_pred - y_train.values

    # create prettierplot object
    p = PrettierPlot(plot_orientation="wide_narrow")

    # add canvas to prettierplot object
    ax = p.make_canvas(
        title="Residual plot - training data\nModel: {}\nParameter set: {}".format(
            model.estimator_name,
            model.model_iter,
        ),
        x_label="Predicted values",
        y_label="Residuals",
        y_shift=0.55,
        title_scale=title_scale,
        position=121,
    )

    # dynamically size precision of x-units based on magnitude of maximum
    # predicted values
    if -1 <= np.nanmax(y_pred) <= 1:
        x_units = "fff"
    elif -100 <= np.nanmax(y_pred) <= 100:
        x_units = "ff"
    else:
        x_units = "f"

    # dynamically size precision of y-units based on magnitude of maximum
    # predicted values
    if -0.1 <= np.nanmax(residuals) <= 0.1:
        y_units = "ffff"
    elif -1 <= np.nanmax(residuals) <= 1:
        y_units = "fff"
    elif -10 <= np.nanmax(residuals) <= 10:
        y_units = "ff"
    else:
        y_units = "f"

    # x tick label rotation
    if -10000 < np.nanmax(y_pred) < 10000:
        x_rotate = 0
    else:
        x_rotate = 45

    # add 2-dimensional scatter plot to canvas
    p.scatter_2d(
        x=y_pred,
        y=residuals,
        size=7,
        color=style.style_grey,
        y_units=y_units,
        x_units=x_units,
        ax=ax,
    )

    # plot horizontal line at y=0
    plt.hlines(
        y=0, xmin=np.min(y_pred), xmax=np.max(y_pred), color=style.style_grey, lw=2
    )

    # add canvas to prettierplot object
    ax = p.make_canvas(
        title="Residual distribution - training data\nModel: {}\nParameter set: {}".format(
            model.estimator_name,
            model.model_iter,
        ),
        title_scale=title_scale,
        position=122,
    )

    # add distribution plot to canvas
    p.dist_plot(
        residuals,
        fit=stats.norm,
        color=style.style_grey,
        y_units="ff",
        x_units="fff",
        ax=ax,
    )
    plt.show()

    # generate regression_stats using training data and predictions
    results = self.regression_stats(
        model=model,
        y_true=y_train.values,
        y_pred=y_pred,
        feature_count=X_train.shape[1],
    )

    # create shell results DataFrame and append
    regression_results_summary = pd.DataFrame(columns=list(results.keys()))
    regression_results_summary = regression_results_summary.append(
        results, ignore_index=True
    )

    ## validation dataset
    # if validation data is provided...
    if X_valid is not None:
        print("\n" + "*" * 55)
        print("Training data evaluation")

        # generate predictions with validation data and calculate residuals
        y_pred = model.predict(X_valid.values)
        residuals = y_pred - y_valid.values

        # create prettierplot object
        p = PrettierPlot(plot_orientation="wide_narrow")

        # add canvas to prettierplot object
        ax = p.make_canvas(
            title="Residual plot - training data\nModel: {}\nParameter set: {}".format(
                model.estimator_name,
                model.model_iter,
            ),
            x_label="Predicted values",
            y_label="Residuals",
            y_shift=0.55,
            title_scale=title_scale,
            position=121,
        )

        # add 2-dimensional scatter plot to canvas
        p.scatter_2d(
            x=y_pred,
            y=residuals,
            size=7,
            color=style.style_grey,
            y_units=y_units,
            x_units=x_units,
            ax=ax,
        )

        # plot horizontal line at y=0
        plt.hlines(
            y=0, xmin=np.min(y_pred), xmax=np.max(y_pred), color=style.style_grey, lw=2
        )

        # add canvas to prettierplot object
        ax = p.make_canvas(
            title="Residual distribution - training data\nModel: {}\nParameter set: {}".format(
                model.estimator_name,
                model.model_iter,
            ),
            title_scale=title_scale,
            position=122,
        )

        # add distribution plot to canvas
        p.dist_plot(
            residuals,
            fit=stats.norm,
            color=style.style_grey,
            y_units="ff",
            x_units="fff",
            ax=ax,
        )
        plt.show()

        # generate regression_stats using validation data and predictions
        results = self.regression_stats(
            model=model,
            y_true=y_valid.values,
            y_pred=y_pred,
            feature_count=X_train.shape[1],
            data_type="validation",
        )

        # append results to regression_results_summary
        regression_results_summary = regression_results_summary.append(
            results, ignore_index=True
        )
        display(regression_results_summary)

    # if n_folds are provided, indicating cross-validation
    elif isinstance(n_folds, int):

        # generate cross-validation indices
        cv = list(
            KFold(
                n_splits=n_folds, shuffle=True, random_state=random_state
            ).split(X_train, y_train)
        )

        print("\n" + "*" * 55)
        print("Cross validation evaluation")

        # iterate through cross-validation indices
        for i, (train_ix, valid_ix) in enumerate(cv):
            X_train_cv = X_train.iloc[train_ix]
            y_train_cv = y_train.iloc[train_ix]
            X_valid_cv = X_train.iloc[valid_ix]
            y_valid_cv = y_train.iloc[valid_ix]

            # fit model on training data and generate predictions using holdout observations
            y_pred = model.fit(X_train_cv.values, y_train_cv.values).predict(
                X_valid_cv.values
            )

            # calculate residuals
            residuals = y_pred - y_valid_cv.values

            # create prettierplot object
            p = PrettierPlot(plot_orientation="wide_narrow")

            # add canvas to prettierplot object
            ax = p.make_canvas(
                title="Residual plot - CV fold {}\nModel: {}\nParameter set: {}".format(
                    i + 1,
                    model.estimator_name,
                    model.model_iter,
                ),
                x_label="Predicted values",
                y_label="Residuals",
                y_shift=0.55,
                position=121,
                title_scale=title_scale,
            )

            # add 2-dimensional scatter plot to canvas
            p.scatter_2d(
                x=y_pred,
                y=residuals,
                size=7,
                color=style.style_grey,
                # color=color_list[i],
                y_units=y_units,
                x_units=x_units,
                ax=ax,
            )

            # plot horizontal line at y=0
            plt.hlines(
                y=0,
                xmin=np.min(y_pred),
                xmax=np.max(y_pred),
                color=style.style_grey,
                lw=2,
            )

            # add canvas to prettierplot object
            ax = p.make_canvas(
                title="Residual distribution - CV fold {}\nModel: {}\nParameter set: {}".format(
                    i + 1,
                    model.estimator_name,
                    model.model_iter,
                ),
                title_scale=title_scale,
                position=122,
            )

            # add distribution plot to canvas
            p.dist_plot(
                residuals,
                fit=stats.norm,
                color=style.style_grey,
                y_units="ff",
                x_units="fff",
                ax=ax,
            )
            plt.show()

            # generate regression_stats using holdout observations and predictions
            results = self.regression_stats(
                model=model,
                y_true=y_valid_cv,
                y_pred=y_pred,
                feature_count=X_valid_cv.shape[1],
                data_type="validation",
                fold=i + 1,
            )

            # append results to regression_results_summary
            regression_results_summary = regression_results_summary.append(
                results, ignore_index=True
            )
        print("\n" + "*" * 55)
        print("Summary")

        display(regression_results_summary)
    else:
        display(regression_results_summary)



import numpy as np
import pandas as pd

from sklearn.metrics import (
    precision_score,
    recall_score,
    f1_score,
    explained_variance_score,
    mean_squared_log_error,
    mean_absolute_error,
    median_absolute_error,
    mean_squared_error,
    r2_score,
    confusion_matrix,
    roc_curve,
    accuracy_score,
    roc_auc_score,
    homogeneity_score,
    completeness_score,
    classification_report,
    silhouette_samples,
)
from sklearn.model_selection import (
    KFold,
    train_test_split,
    GridSearchCV,
    StratifiedKFold,
    cross_val_score,
    RandomizedSearchCV,
)



def top_bayes_optim_models(self, bayes_optim_summary, num_models=1):
    """
    Documentation:

        ---
        Description:
            Aggregate best model(s) for each estimator as determined by bayesian optimization
            hyperparamter tuning process.

        ---
        Paramaters:
            bayes_optim_summary : Pandas DataFrame
                Pandas DataFrame containing results from bayesian optimization process.
            num_models : int, default=1
                Number of top models to return per estimator.

        ---
        Returns:
            models : dictionary
                Dictionary containing string: list key/value pairs, where the string is the
                name of an estimator and the list contains the integer(s) associated with
                the best model(s) as identified in the hyperparameter optimization summary.
    """

    models = {}

    # iterate through unique estimators in bayes_optim_summary
    for estimator in bayes_optim_summary["estimator"].unique():

        # subset bayes_optim_summary by current estimator and keep N rows, where
        # N = num_models. sort by loss and capture only "iteration" values
        est_df = bayes_optim_summary[
            bayes_optim_summary["estimator"] == estimator
        ].sort_values(
            ["loss", "std_score", "train_time"], ascending=[True, True, True]
        )[
            "iteration"
        ][
            :num_models
        ]
        models[estimator] = est_df.values.tolist()
    return models

def binary_prediction_summary(self, model, X_train, y_train, X_valid=None, y_valid=None):
    """
    Documentation:

        ---
        Description:
            Creates a Pandas DataFrame where each row corresponds to an observation, the model's prediction
            for that observation, the true label, and the probabilities associated with the prediction.

        ---
        Paramaters:
            model : model object
                Instantiated model object.
            X_train : Pandas DataFrame
                Training data observations.
            y_train : Pandas Series
                Training target data.
            X_valid : Pandas DataFrame, default=None
                Validation data observations.
            y_valid : Pandas Series, default=None
                Validation target data.

        ---
        Returns:
            df : Pandas DataFrame
                Pandas DataFrame containing prediction summary data.
    """
    # fit model on training data
    model.fit(X_train.values, y_train.values)

    # if no validation data is provided
    if X_valid is None:

        # capture probabilites based on training data
        probas = model.predict_proba(X_train)

        # organize probabilities in dictionary
        data = {
            "Label": y_train,
            "Prediction": probas.argmax(axis=1),
            "Positive": probas[:,0],
            "Negative": probas[:,1],
        }

        # capture data in a DataFrame
        df = pd.DataFrame(data, index=y_train.index)

        # add "Difference" and "Incorrect" summary columns
        df["Difference"] = np.abs(df["Positive"] - df["Negative"])
        df["Incorrect"] = np.abs(df["Label"] - df["Prediction"])

    else:
        # capture probabilites based on validation data
        probas = model.predict_proba(X_valid)

        # organize probabilities in dictionary
        data = {
            "Label": y_valid,
            "Prediction": probas.argmax(axis=1),
            "Positive": probas[:,0],
            "Negative": probas[:,1],
        }

        # capture data in a DataFrame
        df = pd.DataFrame(data, index=y_valid.index)

        # add "Difference" and "Incorrect" summary columns
        df["Difference"] = np.abs(df["Positive"] - df["Negative"])
        df["Incorrect"] = np.abs(df["Label"] - df["Prediction"])

    # sort to bring largest errors to the top
    df = df.sort_values(["Incorrect","Difference"], ascending=[False,False])
    return df

def regression_prediction_summary(self, model, X_train, y_train, X_valid=None, y_valid=None):
    """
    Documentation:

        ---
        Description:
            Creates a Pandas DataFrame where each row corresponds to an observation, the model's prediction
            for that observation, and the true label.

        ---
        Paramaters:
            model : model object
                Instantiated model object.
            X_train : Pandas DataFrame
                Training target data.
            y_train : Pandas Series
                Training data labels.
            X_valid : Pandas DataFrame, default=None
                Validation data observations.
            y_valid : Pandas Series, default=None
                Validation target data.

        ---
        Returns:
            df : Pandas DataFrame
                Pandas DataFrame containing prediction summary data.
    """
    # fit model on training data
    model.fit(X_train.values, y_train.values)

    # if no validation data is provided
    if X_valid is None:

        # generate predictions using training data
        preds = model.predict(X_train)

        # organize data in a dictionary
        data = {
            "Label": y_train,
            "Prediction": preds,
        }

        # capture data in a DataFrame
        df = pd.DataFrame(data, index=y_train.index)

        # add "Difference" and "Incorrect" summary columns
        df["Difference"] = np.abs(df["Label"] - df["Prediction"])
        df["Percent difference"] =  ((df['Prediction'] - df['Label']) / df['Label']) * 100

    else:
        # generate predictions using validation data
        preds = model.predict(X_valid)

        # organize data in a dictionary
        data = {
            "Label": y_valid,
            "Prediction": preds,
        }

        # capture data in a DataFrame
        df = pd.DataFrame(data, index=y_valid.index)

        # add "Difference" and "Incorrect" summary columns
        df["Difference"] = np.abs(df["Label"] - df["Prediction"])
        df["Percent difference"] =  ((df['Prediction'] - df['Label']) / df['Label']) * 100

    # sort to bring largest errors to the top
    df = df.sort_values(["Percent difference"], ascending=[False])
    return df

def regression_stats(self, model, y_true, y_pred, feature_count, fold=0, data_type="training"):
    """
    Documentation:

        ---
        Description:
            Create a dictionary containing information regarding model information and various metrics
            describing the model's performance.

        ---
        Paramaters:
            model : model object
                Instantiated model object.
            y_true : Pandas Series or array
                True labels.
            y_pred : Pandas Series or array
                Predicted labels.
            feature_count : int
                Number of features in the observation data. Used to calculate adjusted R-squared.
            fold : int, default=0
                Indicator for which cross-validation fold the performance is associated with. If 0,
                it is assumed that the evaluation is on an entire dataset (either the full training
                dataset or the full validation dataset), as opposed to a fold.
            data_type : str, default="training"
                String describing type of dataset

        ---
        Returns:
            results : dictionary
                Dictionary containing string: float key/value pairs, where the string is the metric name
                and the value is the metric score

    """
    results = {}

    # model information
    results["Estimator"] = model.estimator_name
    results["Parameter set"] = model.model_iter
    results["Dataset type"] = data_type
    results["CV fold"] = fold
    results["N"] = len(y_true)

    # model performance
    results["Explained Variance"] = explained_variance_score(y_true, y_pred)
    results["Mean squared log error"] = mean_squared_log_error(y_true, y_pred)
    results["Mean absolute error"] = mean_absolute_error(y_true, y_pred)
    results["Median absolute error"] = median_absolute_error(y_true, y_pred)
    results["Mean squared error"] = mean_squared_error(y_true, y_pred)
    results["Root mean squared error"] = np.sqrt(mean_squared_error(y_true, y_pred))
    results["R-squared"] = r2_score(y_true, y_pred)
    results["Adjusted R-squared"] = 1 - (1 - r2_score(y_true, y_pred)) * (
        len(y_true) - 1
    ) / (len(y_true) - feature_count - 1)
    return results

def regression_results(self, model, X_train, y_train, X_valid=None, y_valid=None, n_folds=None,
                        random_state=1, regression_results_summary=None):
    """
    Documentation:

        ---
        Description:
            Creates a Pandas DataFrame where each row captures various summary statistics pertaining to a
            model's performance. Captures performance data for training and validation datasets. If no
            validation set is provided, then cross-validation is performed on the training dataset.

        ---
        Paramaters:
            model : model object
                Instantiated model object.
            X_train : Pandas DataFrame
                Training data observations.
            y_train : Pandas Series
                Training target data.
            X_valid : Pandas DataFrame, default=None
                Validation data observations.
            y_valid : Pandas Series, default=None
                Validation target data.
            n_folds : int, default=None
                Number of cross-validation folds to use when generating
                cv roc graph.
            random_state : int, default=1
                Random number seed.
            regression_results_summary : Pandas DataFrame, default=None
                Pandas DataFrame containing various summary statistics pertaining to model performance. If
                None, returns summary Pandas DataFrame for the input model. If regression_results_summary
                DataFrame is provided from a previous run, the new performance results are appended to the
                provivded summary.

        ---
        Returns:
            regression_results_summary : Pandas DataFrame
                Dataframe containing various summary statistics pertaining to model performance.
    """
    # fit model on training data
    model.fit(X_train.values, y_train.values)

    ## training dataset
    # generate predictions using training data
    y_pred = model.predict(X_train.values)

    # return regression_stats results for training data and predictions
    results = self.regression_stats(
        model=model,
        y_true=y_train.values,
        y_pred=y_pred,
        feature_count=X_train.shape[1],
    )

    # create shell results DataFrame and append
    if regression_results_summary is None:
        regression_results_summary = pd.DataFrame(columns=list(results.keys()))
    regression_results_summary = regression_results_summary.append(
        results, ignore_index=True
    )

    ## validation dataset
    # if validation data is provided...
    if X_valid is not None:

        # generate predictions using validation data
        y_pred = model.predict(X_valid.values)

        # return regression_stats results for validation data and predictions
        results = self.regression_stats(
            model=model,
            y_true=y_valid.values,
            y_pred=y_pred,
            feature_count=X_train.shape[1],
            data_type="validation",
        )

        # append results to DataFrame
        regression_results_summary = regression_results_summary.append(
            results, ignore_index=True
        )
    elif isinstance(n_folds, int):
        # if validation data is not provided, then perform KFold cross-validation on
        # training data
        cv = list(
            KFold(
                n_splits=n_folds, shuffle=True, random_state=random_state
            ).split(X_train, y_train)
        )

        # iterate through KFolds
        for i, (train_ix, valid_ix) in enumerate(cv):
            X_train_cv = X_train.iloc[train_ix]
            y_train_cv = y_train.iloc[train_ix]
            X_valid_cv = X_train.iloc[valid_ix]
            y_valid_cv = y_train.iloc[valid_ix]

            # fit model on training observations and make predictions with holdout observations
            y_pred = model.fit(X_train_cv.values, y_train_cv.values).predict(
                X_valid_cv.values
            )

            # return regression_stats results for holdout data and predictions
            results = self.regression_stats(
                model=model,
                y_true=y_valid_cv,
                y_pred=y_pred,
                feature_count=X_valid_cv.shape[1],
                data_type="validation",
                fold=i + 1,
            )

            # append results to DataFrame
            regression_results_summary = regression_results_summary.append(
                results, ignore_index=True
            )

    return regression_results_summary



import copy
import os
import sys
import importlib
import itertools
import warnings

import numpy as np
import pandas as pd
pd.options.mode.chained_assignment = None

from pandas.api.types import (
    is_bool_dtype,
    is_categorical_dtype,
    is_datetime64_any_dtype,
    is_numeric_dtype,
    is_object_dtype,
    is_string_dtype,
    CategoricalDtype,
)

from sklearn.model_selection import (
    KFold,
    train_test_split,
    GridSearchCV,
    StratifiedKFold,
    cross_val_score,
    RandomizedSearchCV,
)
from sklearn.preprocessing import (
    StandardScaler,
    RobustScaler,
    PolynomialFeatures,
    OrdinalEncoder,
    LabelEncoder,
    OneHotEncoder,
    KBinsDiscretizer,
    QuantileTransformer,
    PowerTransformer,
    MinMaxScaler,
)


class Machine:
    """
    Documentation:
        Description:
            machine facilitates rapid machine learning experimentation tasks, including data
            cleaning, feature encoding, exploratory data analysis, data prepation, model building,
            model tuning and model evaluation.
    """

    # import mlmachine submodules
    from .explore.eda_suite import (
        df_side_by_side,
        eda_cat_target_cat_feat,
        eda_cat_target_num_feat,
        eda_num_target_cat_feat,
        eda_num_target_num_feat,
    )
    from .explore.eda_preprocessing import (
        eda_missing_summary,
        eda_skew_summary,
        eda_transform_box_cox,
        eda_transform_target,
        eda_transform_log1,
    )
    from .features.preprocessing import (
        GroupbyImputer,
        DataFrameSelector,
        DualTransformer,
        KFoldEncoder,
        PandasFeatureUnion,
        PandasTransformer,
        unique_category_levels,
        compare_train_valid_levels,
        missing_column_compare,
        missing_summary,
        skew_summary,
    )
    from .features.outlier import (
        ExtendedIsoForest,
        OutlierIQR,
        outlier_summary,
        outlier_IQR,
    )
    from .features.selection import FeatureSelector
    from .model.evaluate.summarize import (
        binary_prediction_summary,
        regression_prediction_summary,
        regression_results,
        regression_stats,
        top_bayes_optim_models,
    )
    from .model.evaluate.visualize import (
        binary_classification_panel,
        regression_panel,
    )
    from .model.explain.visualize import (
        multi_shap_value_tree,
        multi_shap_viz_tree,
        shap_dependence_grid,
        shap_dependence_plot,
        shap_summary_plot,
        single_shap_value_tree,
        single_shap_viz_tree,
    )
    from .model.tune.bayesian_optim_search import (
        BayesOptimModelBuilder,
        BayesOptimClassifierBuilder,
        BayesOptimRegressorBuilder,
        BasicClassifierBuilder,
        BasicRegressorBuilder,
        BasicModelBuilder,
        exec_bayes_optim_search,
        model_loss_plot,
        model_param_plot,
        objective,
        sample_plot,
        unpack_bayes_optim_summary,
    )
    from .model.tune.power_grid_search import (
        PowerGridModelBuilder,
        PowerGridSearcher,
    )
    from .model.tune.stack import (
        model_stacker,
        oof_generator,
    )

    def __init__(self, data, remove_features=[], identify_as_boolean=None, identify_as_continuous=None, identify_as_count=None,
                identify_as_date=None, identify_as_nominal=None, identify_as_ordinal=None, ordinal_encodings=None,
                identify_as_string=None, target=None, is_classification=None):
        """
        Documentation:

            ---
            Description:
                __init__ handles initial processing of main data set. Creates DataFrame of independent
                variables, Pandas Series containing dependent variable and a dictionary that categorizes
                features by mlm data type.

            ---
            Parameters:
                data : Pandas DataFrame
                    Input data provided as a Pandas DataFrame.
                remove_features : list, default=[]
                    Features to be completely removed from dataset.
                identify_as_boolean : list, default=None
                    Preidentified boolean features. Columns given boolean dtype.
                identify_as_continuous : list, default=None
                    Preidentified continuous features. Columns given float64 dtype.
                identify_as_count : list, default=None
                    Preidentified count features. Columns given int64 dtype.
                identify_as_date : list, default=None
                    Preidentified date features. Columns given datetime64[ns] dtype.
                identify_as_nominal : list, default=None
                    Preidentified nominal category features. Columns given category dtype.
                identify_as_ordinal : list, default=None
                    Preidentified ordinal category features. Columns given category dtype. If
                    an ordinal_encodings dict is passed, the category column will be given the
                    specified order.
                ordinal_encodings : dict, default=None
                    Dictionary where the key is the ordinal column name provided as a
                    string, and the associated value is a list containing the preferred
                    order of the values.
                identify_as_string : list, default=None
                    Preidentified string features. Columns given string dtype.
                target : list, default=None
                    Name of column containing dependent variable.
                is_classification : boolean, default=None
                    Controls whether Machine is instantiated as a classification object or a
                    regression object.

            ---
            Attributes:
                data : Pandas DataFrame
                    Independent variables returned as a Pandas DataFrame.
                target : Pandas Series
                    Dependent variable returned as a Pandas Series.
        """
        self.remove_features = remove_features
        self.target = data[target].squeeze() if target is not None else None
        self.data = (
            data.drop(self.remove_features + [self.target.name], axis=1)
            if target is not None
            else data.drop(self.remove_features, axis=1)
        )
        self.identify_as_continuous = identify_as_continuous
        self.identify_as_boolean = identify_as_boolean
        self.identify_as_count = identify_as_count
        self.identify_as_date = identify_as_date
        self.identify_as_nominal = identify_as_nominal
        self.identify_as_ordinal = identify_as_ordinal
        self.ordinal_encodings = ordinal_encodings
        self.identify_as_string = identify_as_string
        self.is_classification = is_classification

        if self.is_classification is None:
            raise Exception ("Indicate whether supervised learning problem is classification or not by specifying 'is_classification=True' or 'is_classification=False'")

        if self.identify_as_ordinal is not None and self.ordinal_encodings is None:
            warnings.warn("Recommendation - Ordinal column names passed to 'identify_as_ordinal' variable but, no ordinal encoding instructions pass to 'ordinal_encodings' variable. It is recommended to pass a dictionary containing ordinal column names as keys and lists containing the preferred order of encoding as values", UserWarning)

        # execute method capture_mlm_dtypes
        self.data = PreserveMetaData(self.data)
        self.capture_mlm_dtypes()

        # encode the target column if is_classification == True
        if self.target is not None and self.is_classification:
            self.encode_target()

    def capture_mlm_dtypes(self):
        """
        Documentation:

            --
            Description:
                Determine mlm dtype for each feature. Add determination to mlm_dtypes attribute
                and set Pandas dtype in DataFrame accordingly.
        """
        ### populate mlm_dtypes dictionary with feature type label for each feature
        self.data.mlm_dtypes = {}

        ### boolean
        # mlmachine dtype capture
        if isinstance(self.identify_as_boolean, list):
            self.data.mlm_dtypes["boolean"] = self.identify_as_boolean
        elif not isinstance(self.identify_as_boolean, list) and self.identify_as_boolean is not None:
            raise AttributeError ("Variable passed to identify_as_boolean is not a list. Provide a list of column names, provide None or allow identify_as_boolean to default to None.")
        elif self.identify_as_boolean is None:
            self.data.mlm_dtypes["boolean"] = []

        # Pandas dtype
        for column in self.data.mlm_dtypes["boolean"]:
            self.data[column] = self.data[column].astype("boolean")

        ### nominal category
        # mlmachine dtype capture
        if isinstance(self.identify_as_nominal, list):
            self.data.mlm_dtypes["nominal"] = self.identify_as_nominal
        elif not isinstance(self.identify_as_nominal, list) and self.identify_as_nominal is not None:
            raise AttributeError ("Variable passed to identify_as_nominal is not a list. Provide a list of column names, provide None or allow identify_as_nominal to default to None.")
        elif self.identify_as_nominal is None:
            self.data.mlm_dtypes["nominal"] = []

        # Pandas dtype
        for column in self.data.mlm_dtypes["nominal"]:
            self.data[column] = self.data[column].astype("category")

        ### ordinal category
        # mlmachine dtype capture
        if isinstance(self.identify_as_ordinal, list):
            self.data.mlm_dtypes["ordinal"] = self.identify_as_ordinal
        elif not isinstance(self.identify_as_ordinal, list) and self.identify_as_ordinal is not None:
            raise AttributeError ("Variable passed to identify_as_ordinal is not a list. Provide a list of column names, provide None or allow identify_as_ordinal to default to None.")
        elif isinstance(self.ordinal_encodings, dict):
            self.data.mlm_dtypes["ordinal"] = list(self.ordinal_encodings.keys())
        elif self.identify_as_ordinal is None and self.ordinal_encodings is None:
            self.data.mlm_dtypes["ordinal"] = []

        # Pandas dtype
        if isinstance(self.ordinal_encodings, dict):
            for column, order in self.ordinal_encodings.items():
                category_type = CategoricalDtype(categories=order, ordered=True)
                self.data[column] = self.data[column].astype(category_type)

        for column in self.data.mlm_dtypes["ordinal"]:
            self.data[column] = self.data[column].astype("category")

        ### continuous
        # mlmachine dtype capture
        if isinstance(self.identify_as_continuous, list):
            self.data.mlm_dtypes["continuous"] = self.identify_as_continuous
        elif not isinstance(self.identify_as_continuous, list) and self.identify_as_continuous is not None:
            raise AttributeError ("Variable passed to identify_as_continuous is not a list. Either provider a list of column names, provide None or allow identify_as_continuous to default to None.")
        elif self.identify_as_continuous is None:
            self.data.mlm_dtypes["continuous"] = []

        # Pandas dtype
        for column in self.data.mlm_dtypes["continuous"]:
            self.data[column] = self.data[column].astype("float64")

        ### count
        # mlmachine dtype capture
        if isinstance(self.identify_as_count, list):
            self.data.mlm_dtypes["count"] = self.identify_as_count
        elif not isinstance(self.identify_as_count, list) and self.identify_as_count is not None:
            raise AttributeError ("Variable passed to identify_as_count is not a list. Provide a list of column names, provide None or allow identify_as_count to default to None.")
        elif self.identify_as_count is None:
            self.data.mlm_dtypes["count"] = []

        # Pandas dtype
        for column in self.data.mlm_dtypes["count"]:
            try:
                self.data[column] = self.data[column].astype("int64")
            except ValueError:
                self.data[column] = self.data[column].astype("float64")

        ### string
        # mlmachine dtype capture
        if isinstance(self.identify_as_string, list):
            self.data.mlm_dtypes["string"] = self.identify_as_string
        elif not isinstance(self.identify_as_string, list) and self.identify_as_string is not None:
            raise AttributeError ("Variable passed to identify_as_string is not a list. Provide a list of column names, provide None or allow identify_as_string to default to None.")
        elif self.identify_as_string is None:
            self.data.mlm_dtypes["string"] = []

        # Pandas dtype
        for column in self.data.mlm_dtypes["string"]:
            self.data[column] = self.data[column].astype("string")

        ### date
        # mlmachine dtype capture
        if isinstance(self.identify_as_date, list):
            self.data.mlm_dtypes["date"] = self.identify_as_date
        elif not isinstance(self.identify_as_date, list) and self.identify_as_date is not None:
            raise AttributeError ("Variable passed to identify_as_date is not a list. Provide a list of column names, provide None or allow identify_as_date to default to None.")
        elif self.identify_as_date is None:
            self.data.mlm_dtypes["date"] = []

        # Pandas dtype
        for column in self.data.mlm_dtypes["date"]:
            self.data[column] = self.data[column].astype("datetime64[ns]")

        ### untracked columns
        # compile single list of features that have already been categorized
        tracked_columns = [i for i in sum(self.data.mlm_dtypes.values(), [])]

        # iterate through untracked columns and attempt mlmachine dtype identification
        for column in [i for i in self.data.columns if i not in tracked_columns]:

            # capture column statistics and characteristics
            try:
                value_mean = np.mean(self.data[column].dropna())
                value_std = np.std(self.data[column].dropna())
            except TypeError:
                pass

            # identify how many values in feature are zero or one
            zeros_and_ones = (self.data[column].eq(0) | self.data[column].eq(1)).sum()
            
            # count number of unique values in feature
            unique_values = len(np.unique(self.data[column].dropna()))

            # if feature is detected to have an object or categorical dtype
            if is_object_dtype(self.data[column]) \
                or is_string_dtype(self.data[column]) \
                or is_categorical_dtype(self.data[column]):

                self.data.mlm_dtypes["nominal"].append(column)
                self.data[column] = self.data[column].astype("category")

            # if feature is detected to have a datetime64 dtype
            elif is_datetime64_any_dtype(self.data[column]):

                self.data.mlm_dtypes["date"].append(column)
                self.data[column] = self.data[column].astype("datetime64[ns]")

            # if feature is detected to have a bool dtype
            elif is_bool_dtype(self.data[column]):

                self.data.mlm_dtypes["boolean"].append(column)
                self.data[column] = self.data[column].astype("boolean")

            # if feature is detected to have a string dtype
            elif is_string_dtype(self.data[column]):

                self.data.mlm_dtypes["string"].append(column)
                self.data[column] = self.data[column].astype("string")

            # if feature is detected to have a numeric dtype
            elif is_numeric_dtype(self.data[column]):

                # if the values have a significant spread, assume continuous mlm dtype
                if value_std/value_mean > 2 and value_std > 5:
                    self.data.mlm_dtypes["continuous"].append(column)
                    self.data[column] = self.data[column].astype("float64")

                # if feature contains only 0's and 1's, then assume boolean mlm dtype
                elif self.data[column].astype("float").apply(float.is_integer).all() \
                    and zeros_and_ones == self.data.shape[0]:

                    self.data.mlm_dtypes["boolean"].append(column)
                    self.data[column] = self.data[column].astype("boolean")

                # if feature does not contain only 0's and 1's, then assume count mlm dtype
                elif self.data[column].astype("float").apply(float.is_integer).all() \
                    and zeros_and_ones != self.data.shape[0]:

                    self.data.mlm_dtypes["count"].append(column)
                    self.data[column] = self.data[column].astype("int64")

                # otherwise, set as continuous
                else:
                    self.data.mlm_dtypes["continuous"].append(column)
                    self.data[column] = self.data[column].astype("float64")


            # all else are category
            else:
                self.data.mlm_dtypes["nominal"].append(column)
                self.data[column] = self.data[column].astype("category")

        # sort lists within dictionary
        self.data.mlm_dtypes = {x: sorted(self.data.mlm_dtypes[x]) for x in self.data.mlm_dtypes.keys()}

        # create helper keys that combine all category and numeric mlm dtypes
        self.data.mlm_dtypes["category"] = self.data.mlm_dtypes["boolean"] + self.data.mlm_dtypes["nominal"] + self.data.mlm_dtypes["ordinal"]
        self.data.mlm_dtypes["number"] = self.data.mlm_dtypes["continuous"] + self.data.mlm_dtypes["count"]

    def update_dtypes(self, columns_to_drop=None):
        """
        Documentation:

            ---
            Description:
                Update mlm_dtypes dictionary to include new columns. Ensures new object columns
                in dataset have the dtype "category". Optionally drops specific columns from the dataset
                and mlm_dtypes dictionary.

            ---
            Parameters:
                columns_to_drop : list, default=None
                    Columns to drop from output dataset
        """
        ### updates to mlm_dtypes with new columns and drop any specified removals
        # capture current columns
        current_columns = self.data.columns.tolist()

        # capture current state of mlm_dtypes
        old_mlm_dtypes = copy.deepcopy(self.data.mlm_dtypes)

        # remove any columns listed in columns_to_drop
        if columns_to_drop is not None:
            current_columns = [x for x in current_columns if x not in columns_to_drop]

        # remove any columns_to_drop from mlm_dtypes
        for k in self.data.mlm_dtypes.keys():
            self.data.mlm_dtypes[k] = [x for x in self.data.mlm_dtypes[k] if x in current_columns]

        # remove any columns listed in columns_to_drop from the main dataset
        if columns_to_drop is not None:
            try:
                # preserve mlm_dtypes
                self.meta_mlm_dtypes = self.data.mlm_dtypes
                self.data = self.data.drop(columns_to_drop, axis=1)

                # add back mlm_dtypes
                self.data.mlm_dtypes = self.meta_mlm_dtypes
            except KeyError:
                pass

        ### capture nominal column / value pairs
        self.nominal_column_values = {}
        for column in self.data.mlm_dtypes["nominal"]:
            self.nominal_column_values[column] = list(self.data[column].dropna().unique())

        ### add any currently untracked column to mlm_dtypes and set dtype in main dataset
        # compile single list of features that have already been categorized
        tracked_columns = [i for i in sum(self.data.mlm_dtypes.values(), [])]
        untracked_columns = list(set(current_columns).difference(tracked_columns))

        for column in untracked_columns:

            # capture column statistics
            zeros_and_ones = (self.data[column].eq(0) | self.data[column].eq(1)).sum()
            unique_values = len(np.unique(self.data[column].dropna()))

            try:
                value_mean = np.mean(self.data[column].dropna())
                value_std = np.std(self.data[column].dropna())
            except TypeError:
                value_mean = 0.01
                value_std = 0.01
                pass

            # if column contains non-numeric values, default to nominal mlm_dtype
            try:
                self.data[column].astype("float").apply(float.is_integer).all()
            except ValueError:
                self.data.mlm_dtypes["nominal"].append(column)
                self.data[column] = self.data[column].astype("category")
                continue

            # column dtype is object or categorical, contains only integers, and not just 0's and 1's
            if (is_object_dtype(self.data[column]) or is_categorical_dtype(self.data[column])) \
                and self.data[column].astype("float").apply(float.is_integer).all() \
                and zeros_and_ones != self.data.shape[0]:

                self.data.mlm_dtypes["ordinal"].append(column)

                if pd.api.types.is_numeric_dtype(self.data[column].dtype.categories.dtype):
                    order = sorted(self.data[column].unique())
                    category_type = pd.api.types.CategoricalDtype(categories=order, ordered=True)
                    self.data[column] = self.data[column].astype(category_type)

                    self.ordinal_encodings[column] = order
                else:
                    self.data[column] = self.data[column].astype("category")

                    self.ordinal_encodings[column] = self.data[column].unique()

            # if column name suffix indicates that is is a BoxCox or YeoJohnson transformed column
            elif column.endswith(("_BoxCox","_YeoJohnson")):

                self.data.mlm_dtypes["continuous"].append(column)
                self.data[column] = self.data[column].astype("float64")

            # if column name suffix suggests a continuous feature type
            elif column.endswith(("_target_encoded", "_woe_encoded", "_catboost_encoded")):

                self.data.mlm_dtypes["continuous"].append(column)
                self.data[column] = self.data[column].astype("float64")

            # if column name suffix suggests a count feature type
            elif column.endswith(("_count_encoded")):

                self.data.mlm_dtypes["count"].append(column)
                self.data[column] = self.data[column].astype("int64")

            # if column name suffix suggests an ordinal category feature type
            elif column.endswith(("_ordinal_encoded")):

                self.data.mlm_dtypes["ordinal"].append(column)
                self.data[column] = self.data[column].astype("category")

            # if column name suffix suggests an ordinal category feature type
            elif "_binned_" in column:

                self.data.mlm_dtypes["ordinal"].append(column)
                self.data[column] = self.data[column].astype("category")

            # if column name suffix suggests a nominal category feature type
            elif column.endswith(("_binary_encoded")):

                self.data.mlm_dtypes["nominal"].append(column)
                self.data[column] = self.data[column].astype("category")

            # if column name contains "*", and the name split at "*" returns a list of len == 2, and
            # both parts of the column name are a previously identified continuous or count column
            elif "*" in column \
                and len(column.split("*")) == 2 \
                and (column.split("*")[0] in old_mlm_dtypes["continuous"] or column.split("*")[0] in old_mlm_dtypes["count"]) \
                and (column.split("*")[1] in old_mlm_dtypes["continuous"] or column.split("*")[1] in old_mlm_dtypes["count"]):

                self.data.mlm_dtypes["continuous"].append(column)
                self.data[column] = self.data[column].astype("float64")

            # if column name contains "*", and the name split at "*" returns a list of len == 2, and
            # both parts of the column name are a previously identified continuous or count column
            elif column.endswith(("^2")) \
                and (column.split("^")[0] in old_mlm_dtypes["continuous"] or column.split("^")[0] in old_mlm_dtypes["count"]):

                self.data.mlm_dtypes["continuous"].append(column)
                self.data[column] = self.data[column].astype("float64")

            #
            elif value_std/value_mean > 2 and value_std > 5:

                self.data.mlm_dtypes["continuous"].append(column)
                self.data[column] = self.data[column].astype("float64")

            # if column contains only 0's and 1's, and is not a count column
            elif len(column.split("_")) > 1 \
                and column.split("_")[0] not in old_mlm_dtypes["count"] \
                and self.data[column].astype("float").apply(float.is_integer).all() \
                and zeros_and_ones == self.data.shape[0]:

                self.data.mlm_dtypes["nominal"].append(column)
                self.data[column] = self.data[column].astype("category")

            # if first portion of column name is previously identified as a category column
            elif len(column.split("_")) > 1 \
                and column.split("_")[0] in old_mlm_dtypes["nominal"] \
                and column.split("_")[1] in self.nominal_column_values["category"]:

                self.data.mlm_dtypes["nominal"].append(column)
                self.data[column] = self.data[column].astype("category")

            #
            elif self.data[column].astype("float").apply(float.is_integer).all() \
                and zeros_and_ones != self.data.shape[0]:

                self.data.mlm_dtypes["count"].append(column)
                self.data[column] = self.data[column].astype("int64")

            # all else are nominal
            else:

                self.data.mlm_dtypes["nominal"].append(column)
                self.data[column] = self.data[column].astype("category")

        # preserve mlm_dtypes
        self.meta_mlm_dtypes = self.data.mlm_dtypes

        # sort columns alphabetically by name
        self.data = self.data.sort_index(axis=1)

        # add back mlm_dtypes
        self.data.mlm_dtypes = self.meta_mlm_dtypes

        # add helper key / value pairs
        self.data.mlm_dtypes["category"] = self.data.mlm_dtypes["boolean"] + self.data.mlm_dtypes["nominal"] + self.data.mlm_dtypes["ordinal"]
        self.data.mlm_dtypes["number"] = self.data.mlm_dtypes["continuous"] + self.data.mlm_dtypes["count"]

        self.data.mlm_dtypes = {x:sorted(self.data.mlm_dtypes[x]) for x in self.data.mlm_dtypes.keys()}

    def encode_target(self, reverse=False):
        """
        Documentation:

            ---
            Description:
                Encode categorical target column and store as a Pandas Series, where
                the Series name is the name of the feature in the original dataset.

            ---
            Parameters:
                reverse : bool, default=False
                    Reverses encoding of target variables back to original labels.
        """
        # encode label
        self.le_ = LabelEncoder()

        # store as a named Pandas Series
        self.target = pd.Series(
            self.le_.fit_transform(self.target.values.reshape(-1)),
            name=self.target.name,
            index=self.target.index,
        )

        print(">>> category label encoding\n")
        for orig_lbl, enc_lbl in zip(
            np.sort(self.le_.classes_), np.sort(np.unique(self.target))
        ):
            print("\t{} --> {}".format(orig_lbl, enc_lbl))
        print()

        # reverse the label encoding and overwrite target variable with the original data
        if reverse:
            self.target = self.le_.inverse_transform(self.target)

    def recombine_data(self, data=None, target=None):
        """
        Documentation:

            ---
            Description:
                Helper function for recombining the features in the 'data' attribute
                and the 'target' attribute into one Pandas DataFrame.

            ---
            Parameters:
                data : Pandas DataFrame, default=None
                    Pandas DataFrame containing independent variables. If left as none,
                    the feature dataset provided to Machine during instantiation is used.
                target : Pandas Series, default=None
                    Pandas Series containing dependent target variable. If left as none,
                    the target dataset provided to Machine during instantiation is used.
            ---
            Returns:
                df : Pandas DataFrame
                    Pandas DataFrame containing combined independent and dependent variables.
        """
        # use data provided during instantiation if None
        if data is None:
            data = self.data
        # use target provided during instantiation if None
        if target is None:
            target = self.target

        # merge data and target on index
        df = data.merge(target, left_index=True, right_index=True)
        return df


class PreserveMetaData(pd.DataFrame):

    _metadata = ["mlm_dtypes"]

    @property
    def _constructor(self):
        return PreserveMetaData


def train_test_df_compile(data, target_col, valid_size=0.2, random_state=1):
    """
    Documentation:

        ---
        Description:
            Intakes a single dataset and returns a training dataset and a validation dataset
            stored in separate Pandas DataFrames.

        ---
        Parameters:
            data: Pandas DataFrame or array
                Dataset to be deconstructed into train and test sets.
            target_col : str
                Name of target column in dataset
            valid_size : float, default=0.2
                Proportion of dataset to be set aside as "unseen" test data.
            random_state : int, default=1
                Random number seed

        ---
        Returns:
            df_train : Pandas DataFrame
                Training dataset.
            df_valid : Pandas DataFrame
                Validation dataset.
    """
    if isinstance(data, pd.DataFrame):
        y = data[target_col]
        x = data.drop([target_col], axis=1)

    # perform train/test split
    X_train, X_valid, y_train, y_valid = train_test_split(
        x, y, test_size=valid_size, random_state=1, stratify=y
    )

    # merge training data and associated targets
    df_train = X_train.merge(y_train, left_index=True, right_index=True)
    df_valid = X_valid.merge(y_valid, left_index=True, right_index=True)

    return df_train, df_valid



import ast
import csv
import inspect
import sys
import time
from time import gmtime, strftime
from timeit import default_timer as timer

import matplotlib.pyplot as plt
from matplotlib.patches import Patch
import seaborn as sns

import numpy as np
import pandas as pd

from hyperopt import hp, tpe, Trials, fmin, STATUS_OK
from hyperopt.pyll.stochastic import sample


from sklearn.base import TransformerMixin, BaseEstimator, ClassifierMixin, RegressorMixin
from sklearn.decomposition import PCA, LatentDirichletAllocation
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.ensemble import (
    RandomForestClassifier,
    GradientBoostingClassifier,
    AdaBoostClassifier,
    ExtraTreesClassifier,
    IsolationForest,
)
from sklearn.linear_model import (
    Lasso,
    Ridge,
    ElasticNet,
    LinearRegression,
    LogisticRegression,
    SGDRegressor,
)
from sklearn.kernel_ridge import KernelRidge
from sklearn.model_selection import (
    KFold,
    train_test_split,
    GridSearchCV,
    StratifiedKFold,
    cross_val_score,
    RandomizedSearchCV,
)
from sklearn.naive_bayes import MultinomialNB
from sklearn.neighbors import KNeighborsClassifier, KNeighborsRegressor
from sklearn.svm import SVC, SVR
from sklearn.tree import DecisionTreeRegressor, DecisionTreeClassifier
from sklearn.base import clone

from xgboost import XGBClassifier, XGBRegressor
from lightgbm import LGBMClassifier, LGBMRegressor
import catboost

from prettierplot.plotter import PrettierPlot
from prettierplot import style


def objective(space, results_file, estimator_class, data, target, scoring, n_folds, n_jobs):
    """
    Documentation:

        ---
        Description:
            Customizable objective function to minimize in the bayesian hyper-parameter optimization
            process.

        ---
        Parameters:
            space : dictionary
                Dictionary containg 'parameter: value distribution' key/value pairs. The key specifies
                the parameter of the esto,atpr and the optimization process draws trial values from
                that distribution.
            results_file : str
                File destination for results summary csv.
            estimator_class : str or sklearn api object
                The model to be fit.
            data : array
                Training data observations.
            target : array
                Training target data.
            scoring : str
                Evaluation method for scoring model performance. The following metrics are supported:
                    - "accuracy"
                    - "f1_macro"
                    - "f1_micro"
                    - "neg_mean_squared_error"
                    - "roc_auc"
                    - "root_mean_squared_error"
                    - "root_mean_squared_log_error"
                Note that "root_mean_squared_log_error" is not implemented in sklearn. If
                "root_mean_squared_log_error" is specified, model is optimized using
                "neg_mean_squared_error" and then the square root is taken of the absolute
                value of the results, effectively creating the "root_mean_squared_log_error"
                score to minimize.
            n_folds : int
                Number of folds for cross-validation.
            n_jobs : int
                Number of workers to deploy upon execution, if applicable.

        ---
        Returns:
            results : dictionary
                Dictionary containing details for each individual trial. Details include model type,
                iteration, parameter values, run time, and cross-validation summary statistics.
    """
    global ITERATION
    ITERATION += 1
    start = timer()

    # convert select float params to int
    for param in ["num_leaves", "subsample_for_bin", "min_child_samples"]:
        if param in space.keys():
            space[param] = int(space[param])

    # custom metric handling
    if scoring == "root_mean_squared_error":
        scoring = "neg_mean_squared_error"
        score_transform = "rmse"
    elif scoring == "root_mean_squared_log_error":
        scoring = "neg_mean_squared_log_error"
        score_transform = "rmsle"
    else:
        score_transform = scoring

    # create model, estimator name to be used as column header in DataFrame
    model, estimator_name = model_type_check(estimator=estimator_class, n_jobs=n_jobs, params=space)

    # execute cross-validation scoring
    cv = cross_val_score(
        estimator=model.custom_model,
        X=data,
        y=target,
        verbose=False,
        n_jobs=n_jobs,
        cv=n_folds,
        scoring=scoring,
    )

    # log runtime
    run_time = timer() - start

    # calculate loss based on scoring method
    if score_transform in ["accuracy", "f1_micro", "f1_macro", "roc_auc"]:
        loss = 1 - cv.mean()

        mean_score = cv.mean()
        std_score = cv.std()
        min_score = cv.min()
        max_score = cv.max()
    # negative mean squared error
    elif score_transform in ["neg_mean_squared_error"]:
        loss = np.abs(cv.mean())

        mean_score = cv.mean()
        std_score = cv.std()
        min_score = cv.min()
        max_score = cv.max()
    # root mean squared error
    elif score_transform in ["rmse"]:
        cv = np.sqrt(np.abs(cv))
        loss = np.mean(cv)

        mean_score = cv.mean()
        std_score = cv.std()
        min_score = cv.min()
        max_score = cv.max()
    # root mean squared log error
    elif scoring in ["rmsle"]:
        cv = np.sqrt(np.abs(cv))
        loss = np.mean(cv)

        mean_score = cv.mean()
        std_score = cv.std()
        min_score = cv.min()
        max_score = cv.max()

    # export results to csv
    out_file = results_file
    with open(out_file, "a", newline="") as file:
        writer = csv.writer(file)
        writer.writerow(
            [
                ITERATION,
                estimator_name,
                scoring,
                loss,
                mean_score,
                std_score,
                min_score,
                max_score,
                run_time,
                STATUS_OK,
                space,
            ]
        )

    return {
        "iteration": ITERATION,
        "estimator": estimator_name,
        "scoring": scoring,
        "loss": loss,
        "mean_score": mean_score,
        "std_score": std_score,
        "min_score": min_score,
        "max_score": max_score,
        "train_time": run_time,
        "status": STATUS_OK,
        "params": space,
    }

def exec_bayes_optim_search(self, estimator_parameter_space, data, target, scoring, columns=None, n_folds=3,
                        n_jobs=4, iters=50, show_progressbar=False, results_file=None):
    """
    Documentation:

        ---
        Definition:
            Perform bayesian hyper-parameter optimization across a set of models and parameter value
            distributions.

        ---
        Parameters:
            estimator_parameter_space : dictionary of dictionaries
                Dictionary of nested dictionaries. Outer key is a model, and the corresponding value is
                a dictionary. Each nested dictionary contains 'parameter: value distribution' key/value
                pairs. The inner dictionary key specifies the parameter of the model to be tuned, and the
                value is a distribution of values from which trial values are drawn.
            data : array
                Train data observations.
            target : array
                Training target data.
            scoring : str
                Evaluation method for scoring model performance. The following metrics are supported:
                    - "accuracy"
                    - "f1_macro"
                    - "f1_micro"
                    - "neg_mean_squared_error"
                    - "roc_auc"
                    - "root_mean_squared_error"
                    - "root_mean_squared_log_error"
                Note that "root_mean_squared_log_error" is not implemented in sklearn. If
                "root_mean_squared_log_error" is specified, model is optimized using
                "neg_mean_squared_error" and then the square root is taken of the absolute
                value of the results, effectively creating the "root_mean_squared_log_error"
                score to minimize.
            columns : dict or list, default=None
                If list of feature names is provided, all estimators use the same feature set. If dictionary
                provided, dictionary should contain 'estimator: feature names' key/value pairs. The estimator
                key is a string representing an estimator class, and the associated list is the list of features
                to use with that estimator. The dictionary-approach enables utilization of different features
                sets for each estimator.
            n_folds : int, default=3
                Number of folds for cross-validation.
            n_jobs : int, default - 4
                Number of workers to deploy upon execution, if applicable.
            iters : int, default=50
                Number of iterations to run process.
            show_progressbar : boolean, default=False
                Controls whether to print progress bar to console during training.
            results_file : str, default=None
                file destination for results summary csv. If None, defaults to
                ./bayes_optimization_summary_{data}_{time}.csv.
    """
    # set results logging destination
    if results_file is None:
        results_file = "bayes_optimization_summary_{}_{}.csv".format(
            scoring, strftime("%y%m%d%H%M", gmtime())
        )

    # add file header
    with open(results_file, "w", newline="") as outfile:
        writer = csv.writer(outfile)
        writer.writerow(
            [
                "iteration",
                "estimator",
                "scoring",
                "loss",
                "mean_score",
                "std_score",
                "min_score",
                "max_score",
                "train_time",
                "status",
                "params",
            ]
        )

    # iterate through each estimator
    for estimator_class in estimator_parameter_space.keys():

        # set iteration count to zero
        global ITERATION
        ITERATION = 0

        # establish feature space for hyper-parameter search for current estimator
        space = estimator_parameter_space[estimator_class]

        # conditionally handle input data
        if isinstance(data, pd.core.frame.DataFrame):

            # if column subset is provided as a dictionary
            if isinstance(columns, dict):
                try:
                    input_data = data[columns[estimator_class]]
                except KeyError:
                    input_data = data.copy()

            # if column subset is provided as a list
            if isinstance(columns, list):
                try:
                    input_data = data[columns]
                except KeyError:
                    input_data = data.copy()

            elif columns is None:
                pass

            # use underlying numpy ndarray
            input_data = data.values
        elif isinstance(data, np.ndarray):

            # create a copy of the underlying data array
            input_data = data.copy()
        else:
            raise AttributeError(
                "input data set must be either a Pandas DataFrame or a numpy ndarray"
            )

        # conditionally handle input target
        if isinstance(target, pd.core.series.Series):

            # use underlying numpy ndarray
            input_target = target.values
        elif isinstance(target, np.ndarray):

            # create a copy of the underlying data array
            input_target = target.copy()
        else:
            raise AttributeError(
                "input target must be either a Pandas Series or a numpy ndarray"
            )

        # override default arguments with next estimator_class and the cv parameters
        objective.__defaults__ = (
            results_file,
            estimator_class,
            input_data,
            input_target,
            scoring,
            n_folds,
            n_jobs,
        )

        # run optimization
        if show_progressbar:
            print("\n" + "#" * 100)
            print("\nTuning {0}\n".format(estimator_class))

        # minimize the objective function
        best = fmin(
            fn=objective,
            space=space,
            algo=tpe.suggest,
            max_evals=iters,
            trials=Trials(),
            show_progressbar=show_progressbar,
        )

class BayesOptimClassifierBuilder(ClassifierMixin):
    """
    Documentation:

        ---
        Description:
            Helper class for instantiating a classifier model with a provided parameter set as stored
            on the bayesian optimization log file.

        ---
        Parameters:
            bayes_optim_summary : Pandas DataFrame
                Pandas DataFrame containing results from bayesian optimization process.
            estimator_class : str or sklearn api object
                Name of estimator to build.
            model_iter : int
                Numeric identifier for a specific parameter set used in a training
                iteration.
            n_jobs : int, default=4
                Number of workers to use when training the model. This parameter will be
                ignored if the model does not have this parameter.

        ---
        Returns:
            model : model object
                Model instantiated using parameter set.
    """

    def __init__(self, bayes_optim_summary, estimator_class, model_iter, n_jobs=4):
        self.bayes_optim_summary = bayes_optim_summary
        self.estimator_class = estimator_class
        self.model_iter = model_iter
        self.n_jobs = n_jobs
        self.params = self.bayes_optim_summary[
            (self.bayes_optim_summary["estimator"] == self.estimator_class)
            & (self.bayes_optim_summary["iteration"] == self.model_iter)
        ]["params"].values[0]

        # turn string representation of dictionary into dictionary
        self.params = ast.literal_eval(self.params)

        # convert estimator argument to sklearn api object if needed
        if isinstance(self.estimator_class, str):
            self.estimator_name = self.estimator_class
            self.estimator_class = eval(self.estimator_class)

        # capture available model arguments and set probabily and n_jobs where applicable
        estimator_args = inspect.getfullargspec(self.estimator_class).args

        if "probability" in estimator_args:
            self.params["probability"] = True

        if "n_jobs" in estimator_args:
            self.params["n_jobs"] = self.n_jobs

        # instantiate model
        self.custom_model = self.estimator_class(**self.params)

    def train(self, X_train, y_train):
        self.custom_model.fit(X_train, y_train)

    def predict(self, x):
        return self.custom_model.predict(x)

    def predict_proba(self, x):
        return self.custom_model.predict_proba(x)

    def fit(self, x, y):
        return self.custom_model.fit(x, y)

    def feature_importances_(self, x, y):
        return self.custom_model.fit(x, y).feature_importances_

class BayesOptimRegressorBuilder(RegressorMixin):
    """
    Documentation:

        ---
        Description:
            Helper class for instantiating a regression model with a provided parameter set as stored
            on the bayesian optimization log file.

        ---
        Parameters:
            bayes_optim_summary : Pandas DataFrame
                Pandas DataFrame containing results from bayesian optimization process.
            estimator_class : str or sklearn api object
                Name of estimator to build.
            model_iter : int
                Numeric identifier for a specific parameter set used in a training
                iteration.
            n_jobs : int, default=4
                Number of workers to use when training the model. This parameter will be
                ignored if the model does not have this parameter.

        ---
        Returns:
            model : model object
                Model instantiated using parameter set.
    """

    def __init__(self, bayes_optim_summary, estimator_class, model_iter, n_jobs=4):
        self.bayes_optim_summary = bayes_optim_summary
        self.estimator_class = estimator_class
        self.model_iter = model_iter
        self.n_jobs = n_jobs
        self.params = self.bayes_optim_summary[
            (self.bayes_optim_summary["estimator"] == self.estimator_class)
            & (self.bayes_optim_summary["iteration"] == self.model_iter)
        ]["params"].values[0]

        # turn string representation of dictionary into dictionary
        self.params = ast.literal_eval(self.params)

        # convert estimator argument to sklearn api object if needed
        if isinstance(self.estimator_class, str):
            self.estimator_name = self.estimator_class
            self.estimator_class = eval(self.estimator_class)

        # capture available model arguments and set probabily and n_jobs where applicable
        estimator_args = inspect.getfullargspec(self.estimator_class).args

        if "probability" in estimator_args:
            self.params["probability"] = True

        if "n_jobs" in estimator_args:
            self.params["n_jobs"] = self.n_jobs

        # instantiate model
        self.custom_model = self.estimator_class(**self.params)

    def train(self, X_train, y_train):
        self.custom_model.fit(X_train, y_train)

    def predict(self, x):
        return self.custom_model.predict(x)

    def predict_proba(self, x):
        return self.custom_model.predict_proba(x)

    def fit(self, x, y):
        return self.custom_model.fit(x, y)

    def feature_importances_(self, x, y):
        return self.custom_model.fit(x, y).feature_importances_

class BayesOptimModelBuilder(BaseEstimator):
    """
    Documentation:

        ---
        Description:
            Helper class for instantiating a generic model with a provided parameter set as stored
            on the bayesian optimization log file.

        ---
        Parameters:
            bayes_optim_summary : Pandas DataFrame
                Pandas DataFrame containing results from bayesian optimization process.
            estimator_class : str or sklearn api object
                Name of estimator to build.
            model_iter : int
                Numeric identifier for a specific parameter set used in a training
                iteration.
            n_jobs : int, default=4
                Number of workers to use when training the model. This parameter will be
                ignored if the model does not have this parameter.

        ---
        Returns:
            model : model object
                Model instantiated using parameter set.
    """

    def __init__(self, bayes_optim_summary, estimator_class, model_iter, n_jobs=4):
        self.bayes_optim_summary = bayes_optim_summary
        self.estimator_class = estimator_class
        self.model_iter = model_iter
        self.n_jobs = n_jobs
        self.params = self.bayes_optim_summary[
            (self.bayes_optim_summary["estimator"] == self.estimator_class)
            & (self.bayes_optim_summary["iteration"] == self.model_iter)
        ]["params"].values[0]
        self._estimator_type = "classifier"


        # turn string representation of dictionary into dictionary
        self.params = ast.literal_eval(self.params)

        # convert estimator argument to sklearn api object if needed
        if isinstance(self.estimator_class, str):
            self.estimator_name = self.estimator_class
            self.estimator_class = eval(self.estimator_class)

        # capture available model arguments and set probabily and n_jobs where applicable
        estimator_args = inspect.getfullargspec(self.estimator_class).args

        if "probability" in estimator_args:
            self.params["probability"] = True

        if "n_jobs" in estimator_args:
            self.params["n_jobs"] = self.n_jobs

        # instantiate model
        self.custom_model = self.estimator_class(**self.params)

    def train(self, X_train, y_train):
        self.custom_model.fit(X_train, y_train)

    def predict(self, x):
        return self.custom_model.predict(x)

    def predict_proba(self, x):
        return self.custom_model.predict_proba(x)

    def fit(self, x, y):
        return self.custom_model.fit(x, y)

    def feature_importances_(self, x, y):
        return self.custom_model.fit(x, y).feature_importances_

class BasicRegressorBuilder(RegressorMixin):
    """
    Documentation:

        ---
        Description:
            Helper class for instantiating a regressor model with a provided parameter set.

        ---
        Parameters:
            estimator_class : str or sklearn api object
                Model to instantiate.
            params : dictionary, default=None
                Dictionary containing 'parameter: value' pairs. If no dictionary is provided,
                then an empty dictionary is created by default, which instantiates the model with
                its default parameter values.
            n_jobs : int, default=4
                Number of workers to use when training the model. This parameter will be
                ignored if the model does not have this parameter.

        ---
        Returns:
            model : model object
                Model instantiated using parameter set.
    """

    def __init__(self, estimator_class, params=None, n_jobs=4, random_state=0):

        self.estimator_class = estimator_class
        self.params = {} if params is None else params
        self.n_jobs = n_jobs
        self.random_state = random_state

        # convert estimator argument to sklearn api object if needed
        if isinstance(self.estimator_class, str):
            self.estimator_name = self.estimator_class
            self.estimator_class = eval(self.estimator_class)

        # capture available model arguments and set probabily and n_jobs where applicable
        estimator_args = inspect.getfullargspec(self.estimator_class).args

        if "probability" in estimator_args:
            self.params["probability"] = True

        if "n_jobs" in estimator_args:
            self.params["n_jobs"] = self.n_jobs

        if "random_state" in estimator_args:
            self.params["random_state"] = self.random_state

        # instantiate model
        self.custom_model = self.estimator_class(**self.params)

    def train(self, X_train, y_train):
        self.custom_model.fit(X_train, y_train)

    def predict(self, x):
        return self.custom_model.predict(x)

    def predict_proba(self, x):
        return self.custom_model.predict_proba(x)

    def fit(self, x, y):
        return self.custom_model.fit(x, y)

    def feature_importances_(self, x, y):
        return self.custom_model.fit(x, y).feature_importances_

class BasicClassifierBuilder(ClassifierMixin):
    """
    Documentation:

        ---
        Description:
            Helper class for instantiating a classifier model with a provided parameter set.

        ---
        Parameters:
            estimator_class : str or sklearn api object
                model to instantiate.
            params : dictionary, default=None
                dictionary containing 'parameter : value' pairs. If no dictionary is provided,
                then an empty dictionary is created by default, which instantiates the model with
                its default parameter values.
            n_jobs : int, default=4
                Number of workers to use when training the model. This parameter will be
                ignored if the model does not have this parameter.

        ---
        Returns:
            model : model object
                model instantiated using parameter set. model possesses train, predict, fit
                and feature_importances methods.
    """

    def __init__(self, estimator_class, params=None, n_jobs=4, random_state=0):

        self.estimator_class = estimator_class
        self.params = {} if params is None else params
        self.n_jobs = n_jobs
        self.random_state = random_state

        # convert estimator argument to sklearn api object if needed
        if isinstance(self.estimator_class, str):
            self.estimator_name = self.estimator_class
            self.estimator_class = eval(self.estimator_class)

        # capture available model arguments and set probabily and n_jobs where applicable
        estimator_args = inspect.getfullargspec(self.estimator_class).args

        if "probability" in estimator_args:
            self.params["probability"] = True

        if "n_jobs" in estimator_args:
            self.params["n_jobs"] = self.n_jobs

        if "random_state" in estimator_args:
            self.params["random_state"] = self.random_state

        # instantiate model
        self.custom_model = self.estimator_class(**self.params)

    def train(self, X_train, y_train):
        self.custom_model.fit(X_train, y_train)

    def predict(self, x):
        return self.custom_model.predict(x)

    def predict_proba(self, x):
        return self.custom_model.predict_proba(x)

    def fit(self, x, y):
        return self.custom_model.fit(x, y)

    def feature_importances_(self, x, y):
        return self.custom_model.fit(x, y).feature_importances_

class BasicModelBuilder(BaseEstimator):
    """
    Documentation:

        ---
        Description:
            Helper class for instantiating a generic model with a provided parameter set.

        ---
        Parameters:
            estimator_class : str or sklearn api object
                Model to instantiate.
            params : dictionary, default=None
                Dictionary containing 'parameter : value' pairs. If no dictionary is provided,
                then an empty dictionary is created by default, which instantiates the model with
                its default parameter values.
            n_jobs : int, default=4
                Number of workers to use when training the model. This parameter will be
                ignored if the model does not have this parameter.

        ---
        Returns:
            model : model object
                Model instantiated using parameter set.
    """

    def __init__(self, estimator_class, params=None, n_jobs=4, random_state=0):

        self.estimator_class = estimator_class
        self.params = {} if params is None else params
        self.n_jobs = n_jobs
        self.random_state = random_state

        # convert estimator argument to sklearn api object if estimator passed as str
        if isinstance(self.estimator_class, str):
            self.estimator_name = self.estimator_class
            self.estimator_class = eval(self.estimator_class)
        else:
            self.estimator_name = self.estimator_class.__name__

        # capture available model arguments and set probabily and n_jobs where applicable
        estimator_args = inspect.getfullargspec(self.estimator_class).args

        if "probability" in estimator_args:
            self.params["probability"] = True

        if "n_jobs" in estimator_args:
            self.params["n_jobs"] = self.n_jobs

        if "random_state" in estimator_args:
            self.params["random_state"] = self.random_state

        # instantiate model
        self.custom_model = self.estimator_class(**self.params)

    def train(self, X_train, y_train):
        self.custom_model.fit(X_train, y_train)

    def predict(self, x):
        return self.custom_model.predict(x)

    def predict_proba(self, x):
        return self.custom_model.predict_proba(x)

    def fit(self, x, y):
        return self.custom_model.fit(x, y)

    def feature_importances_(self, x, y):
        return self.custom_model.fit(x, y).feature_importances_

def unpack_bayes_optim_summary(self, bayes_optim_summary, estimator_class):
    """
    Documentation:

        ---
        Definition:
            Unpack bayesian optimization results summary data for a specified estimator
            into a Pandas DataFrame where there is one column for each parameter.

        ---
        Parameters:
            bayes_optim_summary : Pandas DataFrame
                Pandas DataFrame containing results from bayesian optimization process.
            estimator_class : str or sklearn api object
                Name of estimator to build.

        ---
        Returns:
            estimator_param_summary : Pandas DataFrame
                Pandas DataFrame where each row is a record of the parameters used and the
                loss recorded in an iteration.
    """
    # subset bayes_optim_summary to specified estimator_class
    estimator_df = bayes_optim_summary[
        bayes_optim_summary["estimator"] == estimator_class
    ].reset_index(drop=True)

    # create a new dataframe for storing estimator_class's parameters
    estimator_summary = pd.DataFrame(
        columns=list(ast.literal_eval(estimator_df.loc[0, "params"]).keys()),
        index=list(range(len(estimator_df))),
    )

    # append parameter set to estimator_summary
    for i, params in enumerate(estimator_df["params"]):
        estimator_summary.loc[i, :] = list(ast.literal_eval(params).values())

    # add columns for loss and iteration number
    estimator_summary["iter_loss"] = estimator_df["loss"]
    estimator_summary["iteration"] = estimator_df["iteration"]

    return estimator_summary

def model_loss_plot(self, bayes_optim_summary, estimator_class, chart_scale=15, trim_outliers=True, outlier_control=1.5,
                    title_scale=0.7, color_map="viridis"):
    """
    Documentation:

        ---
        Definition:
            Visualize how the bayesian optimization loss changes over time across all iterations.
            Extremely poor results are removed from visualized dataset by two filters.
                1) Loss values worse than [loss mean + (2 x loss standard deviation)]
                2) Loss values worse than [median * outliers_control]. 'outlier_control' is a parameter
                   that can be set during function execution.

        ---
        Parameters:
            bayes_optim_summary : Pandas DataFrame
                Pandas DataFrame containing results from bayesian optimization process.
            estimator_class : str or sklearn api object
                Name of estimator to visualize.
            chart_scale : float, default=15
                Control chart proportions. Higher values scale up size of chart objects, lower
                values scale down size of chart objects.
            trim_outliers : boolean, default=True
                Remove extremely high (poor) results by trimming values where the loss is greater
                than 2 standard deviations away from the mean.
            outlier_control : float: default=1.5
                Controls enforcement of outlier trimming. Value is multiplied by median, and the resulting
                product is the cap placed on loss values. Values higher than this cap will be excluded.
                Lower values of outlier_control apply more extreme filtering to loss values.
            title_scale : float, default=0.7
                Controls the scaling up (higher value) and scaling down (lower value) of the size of
                the main chart title, the x_axis title and the y_axis title.
            color_map : str specifying built-in matplotlib colormap, default="viridis"
                Color map applied to plots.
    """
    # unpack bayes_optim_summary parameters for an estimator_class
    estimator_summary = self.unpack_bayes_optim_summary(
        bayes_optim_summary=bayes_optim_summary, estimator_class=estimator_class
    )

    # apply outlier trimming
    if trim_outliers:
        mean = estimator_summary["iter_loss"].mean()
        median = estimator_summary["iter_loss"].median()
        std = estimator_summary["iter_loss"].std()
        cap = mean + (2.0 * std)
        estimator_summary = estimator_summary[
            (estimator_summary["iter_loss"] < cap)
            & (estimator_summary["iter_loss"] < outlier_control * median)
        ]

    # create color list based on color_map
    color_list = style.color_gen(name=color_map, num=3)

    # create prettierplot object
    p = PrettierPlot(chart_scale=chart_scale)

    # add canvas to prettierplot object
    ax = p.make_canvas(
        title="Loss by iteration - {}".format(estimator_class),
        y_shift=0.8,
        position=111,
        title_scale=title_scale,
    )

    # add regression plot to canvas
    p.reg_plot(
        x="iteration",
        y="iter_loss",
        data=estimator_summary,
        y_units="ffff",
        line_color=color_list[0],
        dot_color=color_list[1],
        alpha=0.6,
        line_width=0.4,
        dot_size=10.0,
        ax=ax,
    )
    plt.show()

def model_param_plot(self, bayes_optim_summary, estimator_class, estimator_parameter_space, n_iter, chart_scale=15,
                    color_map="viridis", title_scale=1.2, show_single_str_params=False):
    """
    Documentation:

        ---
        Definition:
            Visualize hyperparameter optimization over all iterations. Compares theoretical distribution to
            the distribution of values that were actually chosen, and visualizes how parameter value
            selections changes over time.

        ---
        Parameters:
            bayes_optim_summary : Pandas DataFrame
                Pandas DataFrame containing results from bayesian optimization process.
            estimator_class : str or sklearn api object
                Name of estimator to visualize.
            estimator_parameter_space : dictionary of dictionaries
                Dictionary of nested dictionaries. Outer key is an estimator, and the corresponding value is
                a dictionary. Each nested dictionary contains 'parameter: value distribution' key/value
                pairs. The inner dictionary key specifies the parameter of the model to be tuned, and the
                value is a distribution of values from which trial values are drawn.
            n_iter : int
                Number of iterations to draw from theoretical distribution in order to visualize the
                theoretical distribution. Higher number leader to more robust distribution but can take
                considerably longer to create.
            chart_scale : float, default=15
                Controls proportions of visualizations. larger values scale visual up in size, smaller values
                scale visual down in size.
            color_map : str specifying built-in matplotlib colormap, default="viridis"
                Color map applied to plots.
            title_scale : float, default=1.2
                Controls the scaling up (higher value) and scaling down (lower value) of the size of
                the main chart title, the x_axis title and the y_axis title.
            show_single_str_params : boolean, default=False
                Controls whether to display visuals for string attributes where there is only one unique value,
                i.e. there was only one choice for the optimization procedure to choose from during each iteration.
    """
    # unpack bayes_optim_summary parameters for an estimator_class
    estimator_summary = self.unpack_bayes_optim_summary(
        bayes_optim_summary=bayes_optim_summary, estimator_class=estimator_class
    )

    # override None with string representation
    estimator_summary = estimator_summary.replace([None], "None")

    # subset estimator_parameter_space to space for the specified estimator_class
    estimator_space = estimator_parameter_space[estimator_class]

    print("*" * 100)
    print("* {}".format(estimator_class))
    print("*" * 100)

    # iterate through each parameter
    for param in estimator_space.keys():

        # sample from theoretical distribution for n_iters
        theoretical_dist = []
        for _ in range(n_iter):
            theoretical_dist.append(sample(estimator_space)[param])

        ## override None with string representation
        # theoretical distribution
        theoretical_dist = ["none" if v is None else v for v in theoretical_dist]
        theoretical_dist = np.array(theoretical_dist)

        # actual distribution
        actual_dist = estimator_summary[param].tolist()
        actual_dist = ["none" if v is None else v for v in actual_dist]
        actual_dist = np.array(actual_dist)

        # limit estimator_summary to "iteration" and current "param" columns
        actual_iter_df = estimator_summary[["iteration", param]]

        # identify how many values in param column are zero or one
        zeros_and_ones = (actual_iter_df[param].eq(True) | actual_iter_df[param].eq(False)).sum()

        # param column only contains zeros and ones, store string representations of "TRUE" and "FALSE"
        if zeros_and_ones == actual_iter_df.shape[0]:
            actual_iter_df = actual_iter_df.replace({True: "TRUE", False: "FALSE"})

        # if theoreitcal distribution has dtype -- np.bool_, store string representations of "TRUE" and "FALSE"
        if isinstance(theoretical_dist[0], np.bool_):
            theoretical_dist = np.array(["TRUE" if i == True else "FALSE" for i in theoretical_dist.tolist()])

            estimator_summary = estimator_summary.replace([True], "TRUE")
            estimator_summary = estimator_summary.replace([False], "FALSE")

        # if theoretical distribution contains str data, then treat this as an object/category parameter
        if any(isinstance(d, str) for d in theoretical_dist):

            # generate color list for stripplot
            stripplot_color_list = style.color_gen(name=color_map, num=len(actual_iter_df[param].unique()) + 1)

            # generate color list for bar chart
            bar_color_list = style.color_gen(name=color_map, num=3)

            # identify unique values and associated count in theoretical distribution
            unique_vals_theo, unique_counts_theo = np.unique(theoretical_dist, return_counts=True)

            # if theoretical distribution only has one unique value and show_single_str_params is set to True
            if len(unique_vals_theo) > 1 or show_single_str_params:

                # identify unique values and associated count in actual distribution
                unique_vals_actual, unique_counts_actual = np.unique(actual_dist, return_counts=True)

                # store data in DataFrame
                df = pd.DataFrame({"param": unique_vals_actual, "Theorical": unique_counts_theo, "Actual": unique_counts_actual})

                # create prettierplot object
                p = PrettierPlot(chart_scale=chart_scale, plot_orientation = "wide_narrow")

                # add canvas to prettierplot object
                ax = p.make_canvas(
                    title="Selection vs. theoretical distribution\n* {0} - {1}".format(estimator_class, param),
                    y_shift=0.8,
                    position=121,
                    title_scale=title_scale,
                )

                # add faceted bar chart to canvas
                p.facet_cat(
                    df=df,
                    feature="param",
                    color_map=bar_color_list[:-1],
                    # color_map=color_map,
                    box=(1.0, 1.15),
                    alpha=1.0,
                    legend_labels=df.columns[1:].values,
                    x_units=None,
                    ax=ax,
                )

                # add canvas to prettierplot object
                ax = p.make_canvas(
                    title="Selection by iteration\n* {0} - {1}".format(estimator_class, param),
                    y_shift=0.5,
                    position=122,
                    title_scale=title_scale,
                )

                # add stripply to canvas
                sns.stripplot(
                    x="iteration",
                    y=param,
                    data=estimator_summary,
                    jitter=0.3,
                    alpha=1.0,
                    size=0.7 * chart_scale,
                    palette=sns.color_palette(stripplot_color_list[:-1]),
                    ax=ax,
                ).set(xlabel=None, ylabel=None)

                # set tick label font size
                ax.tick_params(axis="both", colors=style.style_grey, labelsize=1.2 * chart_scale)

                plt.show()

        # otherwise treat it as a numeric parameter
        else:
            # cast "iteration" as an int and the param values as float
            convert_dict = {"iteration": int, param: float}
            actual_iter_df = actual_iter_df.astype(convert_dict)

            # create color map
            color_list = style.color_gen(name=color_map, num=3)

            # create prettierplot object
            p = PrettierPlot(chart_scale=chart_scale, plot_orientation = "wide_narrow")

            # add canvas to prettierplot object
            ax = p.make_canvas(
                title="Selection vs. theoretical distribution\n* {0} - {1}".format(estimator_class, param),
                y_shift=0.8,
                position=121,
                title_scale=title_scale,
            )

            # dynamically set x-unit precision based on max value
            if -1.0 <= np.nanmax(theoretical_dist) <= 1.0:
                x_units = "fff"
            elif 1.0 < np.nanmax(theoretical_dist) <= 5.0:
                x_units = "ff"
            elif np.nanmax(theoretical_dist) > 5.0:
                x_units = "f"

            # add kernsel density plot for theoretical distribution to canvas
            p.kde_plot(
                theoretical_dist,
                color=color_list[0],
                y_units="ffff",
                x_units=x_units,
                line_width=0.4,
                bw=0.4,
                ax=ax,
            )

            # add kernsel density plot for actual distribution to canvas
            p.kde_plot(
                actual_dist,
                color=color_list[1],
                y_units="ffff",
                x_units=x_units,
                line_width=0.4,
                bw=0.4,
                ax=ax,
            )

            ## create custom legend
            # create labels
            label_color = {}
            legend_labels = ["Theoretical", "Actual"]
            for ix, i in enumerate(legend_labels):
                label_color[i] = color_list[ix]

            # create legend Patches
            Patches = [Patch(color=v, label=k, alpha=1.0) for k, v in label_color.items()]

            # draw legend
            leg = plt.legend(
                handles=Patches,
                fontsize=1.1 * chart_scale,
                loc="upper right",
                markerscale=0.6 * chart_scale,
                ncol=1,
                bbox_to_anchor=(.95, 1.1),
            )

            # label font color
            for text in leg.get_texts():
                plt.setp(text, color="grey")

            # dynamically set y-unit precision based on max value
            if -1.0 <= np.nanmax(actual_iter_df[param]) <= 1.0:
                y_units = "fff"
            elif 1.0 < np.nanmax(actual_iter_df[param]) <= 5.0:
                y_units = "ff"
            elif np.nanmax(actual_iter_df[param]) > 5.0:
                y_units = "f"

            # add canvas to prettierplot object
            ax = p.make_canvas(
                title="Selection by iteration\n* {0} - {1}".format(estimator_class, param),
                y_shift=0.8,
                position=122,
                title_scale=title_scale,
            )

            # add regression plot to canvas
            p.reg_plot(
                x="iteration",
                y=param,
                data=actual_iter_df,
                y_units=y_units,
                x_units="f",
                line_color=color_list[0],
                line_width=0.4,
                dot_color=color_list[1],
                dot_size=10.0,
                alpha=0.6,
                ax=ax
            )
            plt.show()

def sample_plot(self, sample_space, n_iter, chart_scale=15):
    """
    Documentation:

        ---
        Definition:
            Visualizes a single hyperopt theoretical distribution. Useful for helping to determine a
            distribution to use when setting up hyperopt distribution objects for actual parameter
            tuning.

        ---
        Parameters:
            sample_space : dictionary
                Dictionary of 'param name: hyperopt distribution object' key/value pairs. The name can
                be arbitrarily chosen, and the value is a defined hyperopt distribution.
            n_iter : int
                Number of iterations to draw from theoretical distribution in order to visualize the
                theoretical distribution. Higher number leads to more robust distribution but can take
                considerably longer to create.
            chart_scale : float, default=15
                Controls proportions of visualizations. larger values scale visual up in size, smaller values
                scale visual down in size.

    """
    # iterate through each parameter
    for param in sample_space.keys():

        # sample from theoretical distribution for n_iters
        theoretical_dist = []
        for _ in range(n_iter):
            theoretical_dist.append(sample(sample_space)[param])
        theoretical_dist = np.array(theoretical_dist)

        # create prettierplot object
        p = PrettierPlot(chart_scale=chart_scale)

        # add canvas to prettierplot object
        ax = p.make_canvas(
            title="actual vs. theoretical plot\n* {}".format(param),
            y_shift=0.8,
            position=111,
        )

        # add kernel density plot to canvas
        p.kde_plot(
            theoretical_dist,
            color=style.style_grey,
            y_units="p",
            x_units="fff" if np.nanmax(theoretical_dist) <= 5.0 else "ff",
            ax=ax,
        )

def model_type_check(estimator, n_jobs, params=None):
    """
    Documentation:

        ---
        Description:
            Detects type of estimator and create model object. Also returns
            the name of the estimator.

        ---
        Parameters:
            estimator : str, sklearn api object, or instantiated model
                Estimator to build.
            n_jobs : int, default=1
                Number of workers to deploy upon execution, if applicable. If estimator does not
                have an n_jobs parameter, this is ignored.
    """
    # if estimator is passed as str, eval str to create estimator
    if isinstance(estimator, str):
        estimator = eval(estimator)

    # if estimator is an api object, pass through BsicModelBuilder
    if isinstance(estimator, type) or isinstance(estimator, abc.ABCMeta):
        model = BasicModelBuilder(estimator_class=estimator, n_jobs=n_jobs, params=params)
        estimator_name = model.estimator_name

    # otherwise clone the instantiated model that was passed
    else:
        model = clone(estimator)
        estimator_name = retrieve_variable_name(estimator)

    return model, estimator_name

def retrieve_variable_name(variable):

    for fi in reversed(inspect.stack()):
        names = [var_name for var_name, var_val in fi.frame.f_locals.items() if var_val is variable]
        if len(names) > 0:
            return names[0]



import matplotlib.pyplot as plt

import abc
import inspect
from time import gmtime, strftime

import numpy as np
import pandas as pd

from sklearn.feature_selection import (
    f_classif,
    f_regression,
    VarianceThreshold,
    SelectFromModel,
    SelectKBest,
)
from sklearn.model_selection import (
    KFold,
    train_test_split,
    GridSearchCV,
    StratifiedKFold,
    cross_val_score,
    RandomizedSearchCV,
    cross_validate,
)
from sklearn.ensemble import (
    RandomForestClassifier,
    GradientBoostingClassifier,
    AdaBoostClassifier,
    ExtraTreesClassifier,
    IsolationForest,
    RandomForestRegressor,
    GradientBoostingRegressor,
    ExtraTreesRegressor,
    AdaBoostRegressor,
)
from sklearn.linear_model import (
    Lasso,
    Ridge,
    ElasticNet,
    LinearRegression,
    LogisticRegression,
    SGDRegressor,
)
from sklearn.feature_selection import RFE
from sklearn.kernel_ridge import KernelRidge
from sklearn.naive_bayes import MultinomialNB
from sklearn.neighbors import KNeighborsClassifier, KNeighborsRegressor
from sklearn.svm import SVC, SVR
from sklearn.tree import DecisionTreeRegressor, DecisionTreeClassifier
from sklearn.exceptions import NotFittedError
from sklearn.base import clone
from sklearn.metrics import get_scorer, make_scorer

from xgboost import XGBClassifier, XGBRegressor
from lightgbm import LGBMClassifier, LGBMRegressor
import catboost

from prettierplot.plotter import PrettierPlot
from prettierplot import style

from mlxtend.feature_selection import SequentialFeatureSelector

from ..model.tune.bayesian_optim_search import BasicModelBuilder


class FeatureSelector:
    """
    Documentation:

        ---
        Description:
            Evaluate feature importance using several different feature selection techniques,
            including f_score, variance, recursive feature elimination, forward sequential selection,
            backward sequential selection, and correlation to target. Iterates through a list of
            estimators for each technique where applicable. Also includes methods for performing
            cross-validation-driven feature selection and visualization of the results.

        ---
        Parameters:
            data : Pandas DataFrame, default=None
                Pandas DataFrame containing independent variables.
            target : Pandas Series, default=None
                Pandas Series containing  target variable.
            estimators : list of strings, sklearn api objects or instantiated models
                List of estimators to cycle through when executing relevant feature importance
                techniques.
            classification : bool, default=True
                Conditional controlling whether FeatureSelector operates in classification mode
                or regression mode.
    """

    def __init__(self, data, target, estimators, classification=True):
        self.data = data
        self.target = target
        self.estimators = estimators
        self.classification = classification

    def feature_selector_suite(self, sequential_scoring=None, sequential_n_folds=0, rank=False, add_stats=False,
                                n_jobs=1, save_to_csv=False, run_variance=True, run_importance=True, run_rfe=True,
                                run_corr=True, run_f_score=True, run_sfs=True, run_sbs=True):
        """
        Documentation:

            ---
            Description:
                Run multiple feature selections processes and aggregate results. Optionally rank feature
                importance values. Optionally summary statistics on ranks.

            ---
            Parameters:
                sequential_scoring : str, sklearn scoring function, or list of these objects, default=None
                    Scoring metric for sequential feature selector algorithms. If list is provided,
                    algorithm is run for each scoring metric. If None, accuracy is used for classifiers
                    and r2 is used for regressors.
                sequential_n_folds : int, default=0
                    Number of folds to use in cross-validation procedure for sequential feature selector
                    algorithms. If 0 is passed, no cross-validation is performed.
                rank : bool, default=False
                    Conditional controlling whether to overwrite feature importance values with feature
                    importance rank.
                add_stats : bool, default=False
                    Add row-wise summary statistics for feature importance ranking columns. Requires feature
                    importance values to be represented as ranks. If raw feature importance values are provided,
                    values are automatically converted to ranks.
                n_jobs : int, default=1
                    Number of workers to deploy upon execution, if applicable. If estimator does not
                    have an n_jobs parameter, this is ignored.
                save_to_csv : bool, default=Fakse
                    Conditional controlling whether or not the feature selection summary results
                    are saved to a csv file.
                run_variance : bool, default=True
                    Conditional controlling whether or not the variance method is executed.
                run_importance : bool, default=True
                    Conditional controlling whether or not the feature importance method is executed.
                run_rfe : bool, default=True
                    Conditional controlling whether or not the recursive feature elimination method
                    is executed.
                run_corr : bool, default=True
                    Conditional controlling whether or not the correlation to target method is executed.
                run_f_score : bool, default=True
                    Conditional controlling whether or not the F-score method is executed.
                run_sfs : bool, default=True
                    Conditional controlling whether or not the sequential forward feature selection method
                    is executed.
                run_sbs : bool, default=True
                    Conditional controlling whether or not the sequential backward feature selection method
                    is executed.
        """
        # run each feature importance method, if method's parameter is not set to False
        self.results_variance = self.feature_selector_variance(rank=rank) if run_variance else None
        self.results_importance = self.feature_selector_importance(rank=rank, n_jobs=n_jobs) if run_importance else None
        self.results_rfe = self.feature_selector_rfe(n_jobs=n_jobs) if run_rfe else None
        self.results_forward = self.feature_selector_forward_sequential(
                                                            n_jobs=n_jobs,
                                                            scoring=sequential_scoring,
                                                            n_folds=sequential_n_folds,
                                                        ) if run_sfs else None
        self.results_backward = self.feature_selector_backward_sequential(
                                                            n_jobs=n_jobs,
                                                            scoring=sequential_scoring,
                                                            n_folds=sequential_n_folds,
                                                        ) if run_sbs else None
        self.results_corr = self.feature_selector_corr(rank=rank) if run_corr else None
        if self.classification:
            self.results_f_score = self.feature_selector_f_score_class(rank=rank) if run_f_score else None
        else:
            self.results_f_score = self.feature_selector_f_score_reg(rank=rank) if run_f_score else None

        # collect resulting DataFrames in a list
        results = [
            self.results_f_score,
            self.results_variance,
            self.results_corr,
            self.results_rfe,
            self.results_importance,
            self.results_forward,
            self.results_backward,
        ]

        # remove any None values from the DataFrame list
        results = [result for result in results if result is not None]

        # concatenate feature importance summaries
        feature_selector_summary = pd.concat(results, join="inner", axis=1)

        # optionally add summary statistics columns
        if add_stats:
            feature_selector_summary = self.feature_selector_stats(feature_selector_summary)

        # optionally save feature_selector_summary to csv
        if save_to_csv:
            feature_selector_summary.to_csv(
                "feature_selection_summary_{}.csv".format(
                    strftime("%y%m%d%H%M", gmtime())
                ),
                columns=feature_selector_summary.columns,
            )
        return feature_selector_summary

    def feature_selector_f_score_class(self, rank=False):
        """
        Documentation:

            ---
            Description:
                Calculate f-values and p-values for each feature in the context of a
                classification problem.

            ---
            Parameters:
                rank : bool, default=False
                    Conditional controlling whether to overwrite feature importance
                    values with feature importance rank.
        """
        # calculate f-values and p-values
        univariate = f_classif(self.data, self.target)

        # parse data into dictionary
        feature_dict = {}
        feature_dict["f_value"] = univariate[0]
        feature_dict["p_value"] = univariate[1]

        # load data into Pandas DataFrame
        feature_selector_summary = pd.DataFrame(data=feature_dict, index=self.data.columns)

        # optionally overwrite values with rank
        if rank:
            feature_selector_summary = self.apply_ranks(feature_selector_summary)

        return feature_selector_summary

    def feature_selector_f_score_reg(self, rank=False):
        """
        Documentation:

            ---
            Description:
                Calculate f-values and p-values for each feature in the context of a
                regression problem.

            ---
            Parameters:
                rank : bool, default=False
                    Conditional controlling whether to overwrite feature importance values
                    with feature importance rank.
        """
        # calculate f-values and p-values
        univariate = f_regression(self.data, self.target)

        # parse data into dictionary
        feature_dict = {}
        feature_dict["f_value"] = univariate[0]
        feature_dict["p_value"] = univariate[1]

        # load data into Pandas DataFrame
        feature_selector_summary = pd.DataFrame(data=feature_dict, index=self.data.columns)

        # optionally overwrite values with rank
        if rank:
            feature_selector_summary = self.apply_ranks(feature_selector_summary)

        return feature_selector_summary

    def feature_selector_variance(self, rank=False):
        """
        Documentation:

            ---
            Description:
                Calculate variance for each feature.

            --
            Parameters:
                rank : bool, default=False
                    Conditional controlling whether to overwrite feature importance values
                    with feature importance rank.
        """
        # calculate variance
        var_importance = VarianceThreshold()
        var_importance.fit(self.data)

        # load data into Pandas DataFrame
        feature_selector_summary = pd.DataFrame(
            var_importance.variances_,
            index=self.data.columns,
            columns=["variance{}".format("_rank" if rank else "")],
        )

        # optionally overwrite values with rank
        if rank:
            feature_selector_summary = self.apply_ranks(feature_selector_summary)

        return feature_selector_summary

    def feature_selector_importance(self, rank=False, add_stats=False, n_jobs=1):
        """
        Documentation:

            ---
            Description:
                for each estimator, for each feature, calculate feature importance.

            ---
            Parameters:
                rank : bool, default=False
                    Conditional controlling whether to overwrite feature importance values
                    with feature importance rank.
                add_stats : bool, default=False
                    Add row-wise summary statistics for feature importance ranking columns.
                    Requires feature importance values to be represented as ranks. If raw
                    feature importance values are provided, values are automatically converted
                    to ranks.
                n_jobs : int, default=1
                    Number of workers to deploy upon execution, if applicable. If estimator does not
                    have an n_jobs parameter, this is ignored.
        """
        feature_dict = {}

        # iterate through all provided estimators
        for estimator in self.estimators:

            # create model, estimator name to be used as column header in DataFrame
            model, estimator_name = self.model_type_check(estimator=estimator, n_jobs=n_jobs)

            # append suffix to estimator name
            estimator_name =  estimator_name + "_feature_importance"

            ## return feature importance values
            try:
                # if the model has a feature_importances_ attribute
                feature_dict[estimator_name] = model.feature_importances_(self.data.values, self.target)
            except NotFittedError:
                # otherwise, fit model and retrieve resulting feature importance values
                model.fit(self.data.values, self.target)
                feature_dict[estimator_name] = model.feature_importances_
            except AttributeError:
                # otherwise, continue
                continue

        # store data in DataFrame, using column names for the index
        feature_selector_summary = pd.DataFrame(feature_dict, index=self.data.columns)

        # optionally overwrite values with rank
        if rank:
            feature_selector_summary = self.apply_ranks(feature_selector_summary)

        # add summary statistics columns
        if add_stats:
            feature_selector_summary = self.feature_selector_stats(feature_selector_summary)

        return feature_selector_summary

    def feature_selector_rfe(self, add_stats=False, n_jobs=1):
        """
        Documentation:

            ---
            Description:
                For each estimator, perform recursive feature elimination with a step size of 1,
                and use the step in which each feature is removed to describe relative importance.

            ---
            Parameters:
                add_stats : bool, default=False
                    Add row-wise summary statistics for feature importance ranking columns.
                    Requires feature importance values to be represented as ranks. If raw
                    feature importance values are provided, values are automatically converted
                    to ranks.
                n_jobs : int, default=1
                    Number of workers to deploy upon execution, if applicable. If estimator does not
                    have an n_jobs parameter, this is ignored.
        """
        #
        feature_dict = {}

        # iterate through all provided estimators
        for estimator in self.estimators:

            # create model, estimator name to be used as column header in DataFrame
            model, estimator_name = self.model_type_check(estimator=estimator, n_jobs=n_jobs)

            # append suffix to estimator name
            estimator_name =  estimator_name + "_rfe_rank"

            # create exhaustive, single-step recursive feature elimination object
            rfe = RFE(
                    estimator=model.custom_model if hasattr(model, "custom_model") else model,
                    n_features_to_select=1,
                    step=1,
                    verbose=0
                )

            try:
                # fit RFE object and retrieve ranking
                rfe.fit(self.data, self.target)
                feature_dict[estimator_name] = rfe.ranking_
            except (RuntimeError, KeyError):
                continue

        # store data in DataFrame, using column names for the index
        feature_selector_summary = pd.DataFrame(feature_dict, index=self.data.columns)

        # add summary statistics columns
        if add_stats:
            feature_selector_summary = self.feature_selector_stats(feature_selector_summary)

        return feature_selector_summary

    def feature_selector_backward_sequential(self, scoring, n_folds=0, add_stats=False, n_jobs=1):

        """
        Documentation:

            ---
            Description:
                For each estimator, recursively remove features one at a time, capturing
                the step in which each feature is removed.

            ---
            Parameters:
                scoring : str, sklearn scoring function, or list of these objects
                    Scoring metric for sequential feature selector algorithm. If list is provided,
                    algorithm is run for each scoring metric.
                n_folds : int, default=0
                    Number of folds to use in cross-validation procedure. If 0 is passed,
                    no cross-validation is performed.
                add_stats : bool, default=False
                    Add row-wise summary statistics for feature importance ranking columns.
                    Requires feature importance values to be represented as ranks. If raw
                    feature importance values are provided, values are automatically converted
                    to ranks.
                n_jobs : int, default=1
                    Number of workers to deploy upon execution, if applicable. If estimator does not
                    have an n_jobs parameter, this is ignored.
        """
        # if scorers are specified using a list
        if isinstance(scoring, list):
            # if strings, convert to a list of sklearn scorers
            if all(isinstance(metric, str) for metric in scoring):
                scoring = [get_scorer(metric) for metric in scoring]
            # if list of callables, convert to scorers
            elif all(callable(metric) for metric in scoring):
                scoring = [make_scorer(metric) for metric in scoring]

        # if scorer is specified as a single string, convert to a list containing the associated sklearn scorer
        elif isinstance(scoring, str):
            scoring = [get_scorer(scoring)]

        # if single callable is provided, convert to a list contained the associated sklearn scorer
        elif callable(scoring):
            scoring = [make_scorer(scoring)]

        # iterate through estimators and scoring metrics
        results = []

        # iterate through all provided estimators
        for estimator in self.estimators:

            # iterate through all provided scoring techniques
            for metric in scoring:

                # create model, estimator name to be used as column header in DataFrame
                model, estimator_name = self.model_type_check(estimator=estimator, n_jobs=n_jobs)

                # append suffix to estimator name
                try:
                    estimator_name =  estimator_name + "_SBS_rank_" + metric.__name__
                except AttributeError:
                    estimator_name =  estimator_name + "_SBS_rank_" + metric._score_func.__name__

                # create exhaustive sequential feature selector object
                selector = SequentialFeatureSelector(
                            model,
                            k_features=1,
                            forward=False,
                            floating=False,
                            verbose=0,
                            scoring=metric,
                            cv=n_folds,
                            clone_estimator=False,
                    )
                selector = selector.fit(self.data, self.target)

                # collect feature set used in each sequence
                feature_sets = {}
                for index in list(selector.subsets_.keys()):
                    feature_sets[index] = set(selector.subsets_[index]["feature_names"])

                # store data in DataFrame, using column names for the index
                feature_selector_estimator_summary = pd.DataFrame(columns=["feature", estimator_name])

                # iterate through feature count and feature set
                for n_features, feature_set in feature_sets.items():
                    try:

                        # identify column that is in the previous feature set but not in the current feature set
                        selected_feature = list(feature_sets[n_features] - feature_sets[n_features - 1])
                    except:

                        # for final feature set that contain only one feature, simply capture that feature
                        selected_feature = list(feature_sets[n_features])

                    # append the dropped feature with its rank to the feature selector summary DataFrame for the current estimator
                    feature_selector_estimator_summary.loc[len(feature_selector_estimator_summary)] = [selected_feature[0], n_features]
                feature_selector_estimator_summary = feature_selector_estimator_summary.set_index("feature")

                # append estimator summary to list of all results
                results.append(feature_selector_estimator_summary)

        # concatenate all results into single feature selector summary
        feature_selector_summary = pd.concat(results, join="inner", axis=1)
        return feature_selector_summary

    def feature_selector_forward_sequential(self, scoring, n_folds=0, add_stats=False, n_jobs=1):

        """
        Documentation:

            ---
            Description:
                For each estimator, recursively remove features one at a time, capturing
                the step in which each feature is removed.

            ---
            Parameters:
                scoring : str, sklearn scoring function, or list of these objects
                    Scoring metric for sequential feature selector algorithm. If list is provided,
                    algorithm is run for each scoring metric
                n_folds : int, default=0
                    Number of folds to use in cross-validation procedure. If 0 is passed,
                    no cross-validation is performed.
                add_stats : bool, default=False
                    Add row-wise summary statistics for feature importance ranking columns.
                    Requires feature importance values to be represented as ranks. If raw
                    feature importance values are provided, values are automatically converted
                    to ranks.
                n_jobs : int, default=1
                    Number of workers to deploy upon execution, if applicable. If estimator does not
                    have an n_jobs parameter, this is ignored.
        """
        # if scorers are specified using a list
        if isinstance(scoring, list):
            # if strings, convert to a list of sklearn scorers
            if all(isinstance(metric, str) for metric in scoring):
                scoring = [get_scorer(metric) for metric in scoring]
            # if list of callables, convert to scorers
            elif all(callable(metric) for metric in scoring):
                scoring = [make_scorer(metric) for metric in scoring]

        # if scorer is specified as a single string, convert to a list containing the associated sklearn scorer
        elif isinstance(scoring, str):
            scoring = [get_scorer(scoring)]

        # if single callable is provided, convert to a list contained the associated sklearn scorer
        elif callable(scoring):
            scoring = [make_scorer(scoring)]

        # iterate through estimators and scoring metrics
        results = []

        # iterate through all provided estimators
        for estimator in self.estimators:

            # iterate through all provided scoring techniques
            for metric in scoring:

                # create model, estimator name to be used as column header in DataFrame
                model, estimator_name = self.model_type_check(estimator=estimator, n_jobs=n_jobs)

                # append suffix to estimator name
                try:
                    estimator_name =  estimator_name + "_SFS_rank_" + metric.__name__
                except AttributeError:
                    estimator_name =  estimator_name + "_SFS_rank_" + metric._score_func.__name__

                # create exhaustive sequential feature selector object
                selector = SequentialFeatureSelector(
                            model,
                            k_features=self.data.shape[1],
                            forward=True,
                            floating=False,
                            verbose=0,
                            scoring=metric,
                            cv=n_folds,
                            clone_estimator=False,
                    )
                selector = selector.fit(self.data, self.target)

                # collect feature set used in each sequence
                feature_sets = {}
                for index in selector.k_feature_idx_[::-1]:
                    feature_sets[index + 1] = set(selector.subsets_[index + 1]["feature_names"])

                # store data in DataFrame, using column names for the index
                feature_selector_estimator_summary = pd.DataFrame(columns=["feature", estimator_name])

                # iterate through feature count and feature set
                for n_features, feature_set in feature_sets.items():
                    try:

                        # identify column that is in the previous feature set but not in the current feature set
                        selected_feature = list(feature_sets[n_features] - feature_sets[n_features - 1])
                    except:

                        # for final feature set that contain only one feature, simply capture that feature
                        selected_feature = list(feature_sets[n_features])

                    # append the dropped feature with its rank to the feature selector summary DataFrame for the current estimator
                    feature_selector_estimator_summary.loc[len(feature_selector_estimator_summary)] = [selected_feature[0], n_features]
                feature_selector_estimator_summary = feature_selector_estimator_summary.set_index("feature")

                # append estimator summary to list of all results
                results.append(feature_selector_estimator_summary)

        # concatenate all results into single feature selector summary
        feature_selector_summary = pd.concat(results, join="inner", axis=1)
        return feature_selector_summary

    def feature_selector_corr(self, rank=False):
        """
        Documentation:

            ---
            Description:
                For each feature, calculate absolute correlation coefficient relative to
                target dataset.

            ---
            Parameters:
                rank : bool, default=False
                    Conditional controlling whether to overwrite feature importance values with
                    feature importance rank.
        """
        # calculate absolute correlation coefficients relative to target and store in DataFrame
        feature_selector_summary = self.data.merge(self.target, left_index=True, right_index=True)
        feature_selector_summary = pd.DataFrame(feature_selector_summary.corr().abs()[self.target.name])

        # rename, sort correlation coefficient descending, and drop target row
        feature_selector_summary = feature_selector_summary.rename(columns={self.target.name: "correlation_to_target"})
        feature_selector_summary = feature_selector_summary.sort_values(correlation_to_target, ascending=False)
        feature_selector_summary = feature_selector_summary.drop(self.target.name, axis=0)

        # optionally overwrite values with rank
        if rank:
            feature_selector_summary = self.apply_ranks(feature_selector_summary)

        return feature_selector_summary

    def apply_ranks(self, feature_selector_summary):
        """
        Documentation:

            ---
            Description:
                Apply ascending or descending ranking on feature importance values.

            ---
            Parameters:
                feature_selector_summary : Pandas DataFrame
                    Pandas DataFrame with an index corresponding to feature names and columns
                    corresponding to feature importance values.
        """

        # segment feature importance techniques based on how feature importance values translate
        # to relative importance
        ascending_rank = ["p_value","rfe"]
        descending_rank = ["feature_importance","f_value","variance","correlation_to_target"]

        ## ascending
        # capture column to rank in ascending order
        apply_ascending_rank = []
        for column in ascending_rank:
            apply_ascending_rank.extend([s for s in feature_selector_summary.columns if column in s])

        # overwrite values with ascending rank
        for column in apply_ascending_rank:
            feature_selector_summary[column] = feature_selector_summary[column].rank(ascending=True, method="max")
            feature_selector_summary = feature_selector_summary.rename({column: column + "_rank"}, axis=1)

        ## descending
        # capture column to rank in descending order
        apply_descending_rank = []
        for column in descending_rank:
            apply_descending_rank.extend([s for s in feature_selector_summary.columns if column in s])

        # overwrite values with descending rank
        for column in apply_descending_rank:
            feature_selector_summary[column] = feature_selector_summary[column].rank(ascending=False, method="max")
            feature_selector_summary = feature_selector_summary.rename({column: column + "_rank"}, axis=1)
        return feature_selector_summary

    def feature_selector_stats(self, feature_selector_summary):
        """
        Documentation:

            ---
            Description:
                Add row-wise summary statistics for feature importance ranking columns. If data
                provided includes raw feature importance values, ranking will be applied
                automatically.

            ---
            Parameters:
                feature_selector_summary : Pandas DataFrame
                    Pandas DataFrame with an index corresponding to feature names and columns
                    corresponding to feature importance values or feature importance rankings.
        """
        # apply ranking if needed
        if not feature_selector_summary.columns.str.endswith("_rank").sum() == feature_selector_summary.shape[1]:
            feature_selector_summary = self.apply_ranks(feature_selector_summary)

        ## add summary stats
        # mean
        feature_selector_summary.insert(
            loc=0, column="average", value=feature_selector_summary.mean(axis=1)
        )
        # standard deviation
        feature_selector_summary.insert(
            loc=1,
            column="stdev",
            value=feature_selector_summary.iloc[:, 1:].std(axis=1),
        )

        # best rank
        feature_selector_summary.insert(
            loc=2,
            column="best",
            value=feature_selector_summary.iloc[:, 2:].min(axis=1),
        )

        # worst rank
        feature_selector_summary.insert(
            loc=3,
            column="worst",
            value=feature_selector_summary.iloc[:, 3:].max(axis=1),
        )

        # sort on mean rank, ascending
        feature_selector_summary = feature_selector_summary.sort_values("average")
        return feature_selector_summary

    def feature_selector_cross_val(self, scoring, feature_selector_summary, estimators=None, n_folds=3,
                                    step=1, n_jobs=1, verbose=False, save_to_csv=False):
        """
        Documentation:

            ---
            Description:
                Perform cross_validation for each estimator, for progressively smaller sets of features. The list
                of features is reduced by one feature on each pass. The feature removed is the least important
                feature of the remaining set. calculates both the training and test performance.

            ---
            Parameters:
                scoring : str, sklearn scoring function, or list of these objects
                    Scoring metric for cross-validation procedure. If list is provided, procedure is run for
                    each scoring metric.
                feature_selector_summary : Pandas DataFrame or str
                    Pandas DataFrame, or str of csv file location, containing  feature_selector_suite results.
                    If None, use FeatureSelector object's internal attribute.
                estimators : list of strings, sklearn api objects or instantiated models, default=None
                    List of estimators to cycle through when executing relevant feature importance techniques.
                    If None, use estimators provided to FeatureSelector object during instantiation.
                n_folds : int, default=3
                    Number of folds to use in cross-validation procedure.
                step : int, default=1
                    Number of features to remove per iteration.
                n_jobs : int, default=1
                    Number of workers to use when training the model. This parameter will be
                    ignored if the model does not have this parameter.
                verbose : bool, default=False
                    Conditional controlling whether each estimator name is printed prior to initiating
                    cross-validation procedure for that estimator.
                save_to_csv : bool, default=False
                    Conditional controlling whethor or not the cross-validation results are saved to a csv file.
        """
        # load feature selector summary
        if isinstance(feature_selector_summary, str):
            feature_selector_summary = pd.read_csv(
                feature_selector_summary, index_col=0
            )
        elif isinstance(feature_selector_summary, pd.core.frame.DataFrame):
            feature_selector_summary = feature_selector_summary
        else:
            raise AttributeError(
                "No feature_selector_summary detected. Execute one of the feature selector methods or load from .csv"
            )

        # if scorers are specified using a list
        if isinstance(scoring, list):
            # if strings, convert to a list of sklearn scorers
            if all(isinstance(metric, str) for metric in scoring):
                scoring = [get_scorer(metric) for metric in scoring]
            # if list of callables, convert to scorers
            elif all(callable(metric) for metric in scoring):
                scoring = [make_scorer(metric) for metric in scoring]

        # if scorer is specified as a single string, convert to a list containing the associated sklearn scorer
        elif isinstance(scoring, str):
            scoring = [get_scorer(scoring)]

        # if single callable is provided, convert to a list contained the associated sklearn scorer
        elif callable(scoring):
            scoring = [make_scorer(scoring)]

        # add summary stats if needed
        if not "average" in feature_selector_summary.columns:
            feature_selector_summary = self.feature_selector_stats(feature_selector_summary)

        # load estimators if needed
        if estimators is None:
            estimators = self.estimators

        # create empty DataFrame for storing all results
        self.cv_summary = pd.DataFrame(
            columns=[
                "estimator",
                "training score",
                "validation score",
                "scoring",
                "features dropped",
            ]
        )

        # perform cross validation for all estimators for each diminishing set of features
        row_ix = 0

        # iterate through all provided estimators
        for estimator in estimators:

            # optionally print the current estimator's name to track progress
            if verbose:
                print(estimator)

            # create empty DataFrame for storing results for current estimator
            cv = pd.DataFrame(
                columns=[
                    "estimator",
                    "training score",
                    "validation score",
                    "scoring",
                    "features dropped",
                ]
            )

            # create model, estimator name to be used as column header in DataFrame
            model, estimator_name = self.model_type_check(estimator=estimator, n_jobs=n_jobs)

            # iterate through scoring metrics
            for metric in scoring:

                step_increment = 0

                # iterate through each set of features
                for i in np.arange(0, feature_selector_summary.shape[0], step):

                    # collect names of columns to form feature subset
                    if i == 0:
                        top = feature_selector_summary.sort_values("average").index
                    else:
                        top = feature_selector_summary.sort_values("average").index[:-i]

                    # special-case handling of certain metrics and metric names
                    if metric == "root_mean_squared_error":
                        metric = "neg_mean_squared_error"
                        score_transform = "rmse"
                    elif metric == "root_mean_squared_log_error":
                        metric = "neg_mean_squared_log_error"
                        score_transform = "rmsle"
                    else:
                        score_transform = metric

                    # execute cross-validation procedure with current feature subset
                    scores = cross_validate(
                        estimator=model.custom_model if hasattr(model, "custom_model") else model,
                        X=self.data[top],
                        y=self.target,
                        cv=n_folds,
                        scoring=metric,
                        return_train_score=True,
                    )

                    # special-case handling of certain metrics and metric names
                    if score_transform == "rmse":
                        training = np.mean(np.sqrt(np.abs(scores["train_score"])))
                        validation = np.mean(np.sqrt(np.abs(scores["test_score"])))
                        metric_name = "root_mean_squared_error"
                    elif score_transform == "rmsle":
                        training = np.mean(np.sqrt(np.abs(scores["train_score"])))
                        validation = np.mean(np.sqrt(np.abs(scores["test_score"])))
                        metric_name = "root_mean_squared_log_error"
                    else:
                        training = np.mean(scores["train_score"])
                        validation = np.mean(scores["test_score"])
                        metric_name = metric._score_func.__name__

                    # append results to estimator's summary DataFrame
                    cv.loc[row_ix] = [
                        estimator_name,
                        training,
                        validation,
                        metric_name,
                        step_increment,
                    ]
                    step_increment += step
                    row_ix += 1

            # append estimator cross-validation summary to main summary DataFrame
            self.cv_summary = self.cv_summary.append(cv)

        # optionally save results to csv
        if save_to_csv:
            self.cv_summary.to_csv(
                "cv_summary_{}.csv".format(strftime("%y%m%d%H%M", gmtime())),
                columns=self.cv_summary.columns,
                index_label="index",
            )

        return self.cv_summary

    def feature_selector_results_plot(self, scoring, feature_selector_summary, cv_summary=None, top_sets=0,
                                    show_features=False, show_scores=None, marker_on=True, title_scale=0.7,
                                    chart_scale=15):
        """
        Documentation:

            ---
            Description:
                For each estimator, visualize the training and validation performance
                for each feature set.

            ---
            Parameters:
                scoring : str
                    Scoring metric to visualize.
                cv_summary : Pandas DataFrame or str, default=None
                    Pandas DataFrame, or str of csv file location, containing cross-validation results.
                    If None, use FeatureSelector object's internal attribute.
                feature_selector_summary : Pandas DataFrame or str
                    Pandas DataFrame, or str of csv file location.
                top_sets : int, default=0
                    Display performance for the top N feature sets.
                show_features : bool, default=False
                    Conditional controlling whether to print feature set for best validation score.
                show_scores : int or none, default=None
                    Display certain number of top features. If None, display nothing. If int, display
                    the specified number of features as a Pandas DataFrame.
                marker_on : bool, default=True
                    Conditional controlling whether to display marker for each individual score.
                title_scale : float, default=0.7
                    Controls the scaling up (higher value) and scaling down (lower value) of the size of
                    the main chart title, the x_axis title and the y_axis title.
                chart_scale : float or int, default=15
                    Chart proportionality control. Determines relative size of figure size, axis labels,
                    Chart title, tick labels, and tick marks.
        """
        # load feature selector summary
        if isinstance(feature_selector_summary, str):
            feature_selector_summary = pd.read_csv(
                feature_selector_summary, index_col=0
            )
        elif isinstance(feature_selector_summary, pd.core.frame.DataFrame):
            feature_selector_summary = feature_selector_summary
        else:
            raise AttributeError(
                "No feature_selector_summary detected. either execute method feature_selector_suite or load from .csv"
            )

        # load cv_summary if needed
        if isinstance(cv_summary, str):
            cv_summary = pd.read_csv(cv_summary, index_col=0)
        elif isinstance(cv_summary, pd.core.frame.DataFrame):
            cv_summary = cv_summary
        elif cv_summary is None:
            cv_summary = self.cv_summary

        # add summary stats if needed
        if not "average" in feature_selector_summary.columns:
            feature_selector_summary = self.feature_selector_stats(feature_selector_summary)

        # iterate through each unique estimator in cv_summary
        for estimator in cv_summary["estimator"].unique():

            # subset cv_summary by scoring method and current estimator
            cv = cv_summary[
                (cv_summary["scoring"] == scoring)
                & (cv_summary["estimator"] == estimator)
            ]

            # capture total number of features, iterations, and step size
            total_features = feature_selector_summary.shape[0]
            iters = cv.shape[0]
            step = np.ceil(total_features / iters)

            # reset index based on step size
            cv.set_index(
                keys=np.arange(0, cv.shape[0] * step, step, dtype=int), inplace=True
            )

            # optionally display top N scores, where N = show_scores
            if show_scores is not None:
                display(cv[:show_scores])

            ## capture best iteration's feature drop count and performance score
            # special handling for select metrics
            if scoring in ["root_mean_squared_error", "root_mean_squared_log_error"]:
                sort_order = True
            else:
                sort_order = False

            # retrieve number of features dropped to form best subset
            num_dropped = cv.sort_values(["validation score"], ascending=sort_order)[
                :1
            ]["features dropped"].values[0]

            # retrieve score achieved by best subset
            score = np.round(
                cv.sort_values(["validation score"], ascending=sort_order)[
                    "validation score"
                ][:1].values[0],
                5,
            )

            # display performance for the top n feature sets
            if top_sets > 0:
                display(
                    cv.sort_values(["validation score"], ascending=sort_order)[
                        :top_sets
                    ]
                )

            # optionally print feature names of features used in best subset
            if show_features:
                if num_dropped > 0:
                    features = feature_selector_summary.shape[0] - num_dropped
                    features_used = (
                        feature_selector_summary.sort_values("average")
                        .index[:features]
                        .values
                    )
                else:
                    features_used = feature_selector_summary.sort_values(
                        "average"
                    ).index.values
                print(features_used)

            # create prettierplot object
            p = PrettierPlot(chart_scale=chart_scale)

            # add canvas to prettierplot object
            ax = p.make_canvas(
                title="{}\nBest validation {} = {}\nFeatures dropped = {}".format(
                    estimator, scoring, score, num_dropped
                ),
                x_label="features removed",
                y_label=scoring,
                y_shift=0.4 if len(scoring) > 18 else 0.57,
                title_scale=title_scale,
            )

            # plot multi-line plot on canvas
            p.multi_line(
                x=cv.index,
                y=["training score", "validation score"],
                label=["training score", "validation score"],
                df=cv,
                y_units="fff",
                marker_on=marker_on,
                bbox=(1.3, 0.9),
                ax=ax,
            )
            plt.show()

    def create_cross_val_features_df(self, scoring, feature_selector_summary, cv_summary=None):
        """
        Documentation:

            ---
            Description:
                for each estimator, visualize the training and validation performance
                for each feature set.

            ---
            Parameters:
                scoring : str
                    Scoring metric to visualize.
                feature_selector_summary : Pandas DataFrame or str
                    Pandas DataFrame, or str of csv file location, containing summary of
                    feature_selector_suite results.
                cv_summary : Pandas DataFrame or str, default=None
                    Pandas DataFrame, or str of csv file location, containing cross-validation results.
                    If None, use FeatureSelector object's internal attribute.
        """
        # load feature selector summary
        if isinstance(feature_selector_summary, str):
            feature_selector_summary = pd.read_csv(
                feature_selector_summary, index_col=0
            )
        elif isinstance(feature_selector_summary, pd.core.frame.DataFrame):
            feature_selector_summary = feature_selector_summary
        else:
            raise AttributeError(
                "No feature_selector_summary detected. either execute method feature_selector_suite or load from .csv"
            )

        # load cv_summary if needed
        if isinstance(cv_summary, str):
            cv_summary = pd.read_csv(cv_summary, index_col=0)
        elif isinstance(cv_summary, pd.core.frame.DataFrame):
            cv_summary = cv_summary
        elif cv_summary is None:
            cv_summary = self.cv_summary

        # add summary stats if needed
        if not "average" in feature_selector_summary.columns:
            feature_selector_summary = self.feature_selector_stats(feature_selector_summary)

        # create empty DataFrame with feature names as index
        self.cross_val_features_df = pd.DataFrame(index=feature_selector_summary.index)

        # iterate through estimators
        for estimator in cv_summary["estimator"].unique():

            # subset cv_summary by scoring method and current estimator
            cv = cv_summary[
                (cv_summary["scoring"] == scoring)
                & (cv_summary["estimator"] == estimator)
            ]
            cv = cv.reset_index(drop=True)

            ## capture best iteration's feature drop count
            # special handling for select metrics
            if scoring in ["root_mean_squared_error", "root_mean_squared_log_error"]:
                sort_order = True
            else:
                sort_order = False

            # retrieve number of features dropped to form best subset
            num_dropped = cv.sort_values(["validation score"], ascending=sort_order)[
                :1
            ]["features dropped"].values[0]
            if num_dropped > 0:
                features = feature_selector_summary.shape[0] - num_dropped
                features_used = (
                    feature_selector_summary.sort_values("average")
                    .index[:features]
                    .values
                )
            else:
                features_used = feature_selector_summary.sort_values(
                    "average"
                ).index.values

            # create column for estimator and populate with marker
            self.cross_val_features_df[estimator] = np.nan
            self.cross_val_features_df[estimator].loc[features_used] = "x"

        # add counter and fill nans
        self.cross_val_features_df["count"] = self.cross_val_features_df.count(axis=1)
        self.cross_val_features_df = self.cross_val_features_df.fillna("")

        # add numeric index starting at 1
        self.cross_val_features_df = self.cross_val_features_df.reset_index()
        self.cross_val_features_df.index = np.arange(
            1, len(self.cross_val_features_df) + 1
        )
        self.cross_val_features_df = self.cross_val_features_df.rename(
            columns={"index": "feature"}
        )
        return self.cross_val_features_df

    def create_cross_val_features_dict(self, scoring, feature_selector_summary, cv_summary=None):
        """
        Documentation:

            ---
            Description:
                for each estimator, visualize the training and validation performance
                for each feature set.

            ---
            Parameters:
                scoring : str
                    Scoring metric to visualize.
                feature_selector_summary : Pandas DataFrame or str
                    Pandas DataFrame, or str of csv file location, containing summary of feature_selector_suite results.
                    If None, use object's internal attribute specified during instantiation.
                cv_summary : Pandas DataFrame or str, default=None
                    Pandas DataFrame, or str of csv file location, containing cross_validation results.
                    If None, use object's internal attribute specified during instantiation.
        """

        # load feature selector summary
        if isinstance(feature_selector_summary, str):
            feature_selector_summary = pd.read_csv(
                feature_selector_summary, index_col=0
            )
        elif isinstance(feature_selector_summary, pd.core.frame.DataFrame):
            feature_selector_summary = feature_selector_summary
        else:
            raise AttributeError(
                "No feature_selector_summary detected. either execute method feature_selector_suite or load from .csv"
            )

        # load cv_summary if needed
        if isinstance(cv_summary, str):
            cv_summary = pd.read_csv(cv_summary, index_col=0)
        elif isinstance(cv_summary, pd.core.frame.DataFrame):
            cv_summary = cv_summary
        elif cv_summary is None:
            raise AttributeError(
                "no cv_summary detected. either execute method feature_selector_suite or load from .csv"
            )

        # execute create_cross_val_features_df
        cross_val_features_df = self.create_cross_val_features_df(
                                                scoring=scoring,
                                                cv_summary=cv_summary,
                                                feature_selector_summary=feature_selector_summary
                                            )

        # create empty dict with feature names as index
        self.cross_val_features_dict = {}

        cross_val_features_df = cross_val_features_df.set_index("feature")
        cross_val_features_df = cross_val_features_df.drop("count", axis=1)

        # iterate through estimators
        for estimator in cross_val_features_df.columns:

            # filter to rows with an "x"
            self.cross_val_features_dict[estimator] = cross_val_features_df[
                cross_val_features_df[estimator] == "x"
            ][estimator].index

        return self.cross_val_features_dict

    def model_type_check(self, estimator, n_jobs=1):
        """
        Documentation:

            ---
            Description:
                Detects type of estimator and create model object. Also returns
                the name of the estimator.

            ---
            Parameters:
                estimator : str, sklearn api object, or instantiated model
                    Estimator to build.
                n_jobs : int, default=1
                    Number of workers to deploy upon execution, if applicable. If estimator does not
                    have an n_jobs parameter, this is ignored.
        """
        # if estimator is passed as str, eval str to create estimator
        if isinstance(estimator, str):
            estimator = eval(estimator)

        # if estimator is an api object, pass through BasicModelBuilder
        if isinstance(estimator, type) or isinstance(estimator, abc.ABCMeta):
            model = BasicModelBuilder(estimator_class=estimator, n_jobs=n_jobs)
            estimator_name = model.estimator_name

        # otherwise clone the instantiated model that was passed
        else:
            model = clone(estimator)
            estimator_name =  self.retrieve_variable_name(estimator)

        return model, estimator_name

    def retrieve_variable_name(self, variable):

        for fi in reversed(inspect.stack()):
            names = [var_name for var_name, var_val in fi.frame.f_locals.items() if var_val is variable]
            if len(names) > 0:
                return names[0]



import joblib
import os

dir = os.path.dirname(os.path.realpath(__file__))


def attrition():
    """
    Documentation:

        ---
        Description:
            Load IBM Employee Attrition dataset.
    """
    data = joblib.load(os.path.join(dir, "datasets/attrition/data.pkl"))
    return data


def housing():
    """
    Documentation:

        ---
        Description:
            Load Kaggle Housing Prices training dataset and validation dataset.
    """
    train = joblib.load(os.path.join(dir, "datasets/housing//train.pkl"))
    test = joblib.load(os.path.join(dir, "datasets/housing/test.pkl"))
    return train, test


def titanic():
    """
    Documentation:

        ---
        Description:
            Load Kaggle Titanic Survivorship training dataset and validation dataset.
    """
    train = joblib.load(os.path.join(dir, "datasets/titanic/train.pkl"))
    test = joblib.load(os.path.join(dir, "datasets/titanic/test.pkl"))
    return train, test

